# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for download."""

import unittest
import urllib.parse

import flask
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers import download
from libs.issue_management import issue_tracker


@test_utils.with_cloud_emulators('datastore')
class DownloadTest(unittest.TestCase):
  """Test download handler."""

  def setUp(self):
    self.minimized_key = 'minimized'
    self.fuzzed_key = 'fuzzed'
    self.unused_key = 'unused'

    self.testcase = data_types.Testcase(
        minimized_keys=self.minimized_key,
        fuzzed_keys=self.fuzzed_key,
        bug_information='1337')
    self.testcase.put()

    test_helpers.patch(self, [
        'clusterfuzz._internal.base.utils.is_oss_fuzz',
        'clusterfuzz._internal.google_cloud_utils.blobs.get_blob_info',
        'libs.access.can_user_access_testcase',
        'libs.access.has_access',
        'libs.gcs.get_signed_url',
        'libs.helpers.get_user_email',
        'libs.issue_management.issue_tracker_utils.'
        'get_issue_tracker_for_testcase',
    ])
    self.mock.is_oss_fuzz.return_value = False
    self.mock.can_user_access_testcase.return_value = False
    self.mock.has_access.return_value = False
    self.mock.get_user_email.return_value = 'a@user.com'
    self.mock.get_signed_url.side_effect = (
        lambda b, p: 'https://SIGNED_URL?path=' + p)

    self.mock.get_blob_info.side_effect = (
        lambda key: storage.GcsBlobInfo('blobs-bucket', key, 'file.ext', 1337))

    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/download', view_func=download.Handler.as_view(''))
    flaskapp.add_url_rule(
        '/download/<resource>',
        view_func=download.Handler.as_view('/download/<resource>'))
    self.app = webtest.TestApp(flaskapp)

  def _test_download(self,
                     blob_key=None,
                     testcase_id=None,
                     expect_status=302,
                     expect_blob=True,
                     expect_filename=None):
    """Test a download."""
    request = '/download'
    if blob_key:
      request += '/%s' % blob_key
    if testcase_id:
      request += '?testcase_id=%s' % str(testcase_id)

    resp = self.app.get(request, expect_errors=True)
    self.assertEqual(expect_status, resp.status_int)

    expected_blob_key = blob_key if blob_key else self.minimized_key
    expected_url = 'https://SIGNED_URL?path=' + expected_blob_key

    if expect_filename:
      expected_url += (
          '&response-content-disposition='
          'attachment%3B+filename%3D' + urllib.parse.quote(expect_filename))

    if expect_blob:
      self.assertEqual(expected_url, resp.location)
    elif expect_status == 302:
      self.assertNotIn('SIGNED_URL', resp.location)

  def test_download_nothing(self):
    """Test download fails when nothing is requested."""
    self._test_download('', expect_status=400, expect_blob=False)

  def test_download_not_logged_in(self):
    """Test download (not logged in)."""
    self.mock.get_user_email.return_value = ''
    self._test_download(
        self.minimized_key, expect_status=302, expect_blob=False)
    self._test_download(self.fuzzed_key, expect_status=302, expect_blob=False)
    self._test_download(self.unused_key, expect_status=302, expect_blob=False)

    self._test_download(
        testcase_id=self.testcase.key.id(),
        expect_status=302,
        expect_blob=False)
    self._test_download(
        self.minimized_key,
        testcase_id=self.testcase.key.id(),
        expect_status=302,
        expect_blob=False)
    self._test_download(
        self.fuzzed_key,
        testcase_id=self.testcase.key.id(),
        expect_status=302,
        expect_blob=False)

  def test_download_user(self):
    """Test download with an internal user."""
    self.mock.has_access.return_value = True
    self._test_download(self.minimized_key, expect_filename='file.ext')
    self._test_download(self.fuzzed_key, expect_filename='file.ext')
    self._test_download(self.unused_key, expect_filename='file.ext')

    self._test_download(
        testcase_id=self.testcase.key.id(),
        expect_filename='clusterfuzz-testcase-minimized-1.ext')
    self._test_download(
        self.minimized_key,
        testcase_id=self.testcase.key.id(),
        expect_filename='clusterfuzz-testcase-minimized-1.ext')
    self._test_download(
        self.fuzzed_key,
        testcase_id=self.testcase.key.id(),
        expect_filename='clusterfuzz-testcase-1.ext')

  def test_download_testcase_user(self):
    """Test download with an user that has access to a testcase."""
    self.mock.can_user_access_testcase.return_value = True
    self._test_download(
        self.minimized_key, expect_status=403, expect_blob=False)
    self._test_download(self.fuzzed_key, expect_status=403, expect_blob=False)
    self._test_download(self.unused_key, expect_status=403, expect_blob=False)

    self._test_download(
        testcase_id=self.testcase.key.id(),
        expect_filename='clusterfuzz-testcase-minimized-1.ext')
    self._test_download(
        self.minimized_key,
        testcase_id=self.testcase.key.id(),
        expect_filename='clusterfuzz-testcase-minimized-1.ext')
    self._test_download(
        self.fuzzed_key,
        testcase_id=self.testcase.key.id(),
        expect_filename='clusterfuzz-testcase-1.ext')

  def test_download_testcase_no_user_access(self):
    """Test download testcase with a user that doesn't have access to the
    testcase."""
    self.mock.can_user_access_testcase.return_value = False
    self._test_download(
        testcase_id=self.testcase.key.id(),
        expect_status=403,
        expect_blob=False)
    self._test_download(
        self.minimized_key,
        testcase_id=self.testcase.key.id(),
        expect_status=403,
        expect_blob=False)
    self._test_download(
        self.fuzzed_key,
        testcase_id=self.testcase.key.id(),
        expect_status=403,
        expect_blob=False)

  def test_public_download_chromium(self):
    """Test public downloading chromium testcases (should fail)."""
    self.mock.get_user_email.return_value = ''
    self.mock.is_oss_fuzz.return_value = False
    mock_issue = self.mock.get_issue_tracker_for_testcase(None).get_issue()
    mock_issue.labels = issue_tracker.LabelStore([])

    self._test_download(
        self.minimized_key, expect_status=302, expect_blob=False)
    self._test_download(
        testcase_id=self.testcase.key.id(),
        expect_status=302,
        expect_blob=False)
    self._test_download(
        self.minimized_key,
        testcase_id=self.testcase.key.id(),
        expect_status=302,
        expect_blob=False)
    self._test_download(
        self.fuzzed_key,
        testcase_id=self.testcase.key.id(),
        expect_status=302,
        expect_blob=False)

  def test_public_download_oss_fuzz(self):
    """Test public downloading OSS-Fuzz testcases."""
    self.mock.get_user_email.return_value = ''

    self.mock.is_oss_fuzz.return_value = True
    mock_issue = self.mock.get_issue_tracker_for_testcase(None).get_issue()
    mock_issue.labels = issue_tracker.LabelStore([])

    self._test_download(
        self.minimized_key, expect_status=302, expect_blob=False)
    self._test_download(
        testcase_id=self.testcase.key.id(),
        expect_filename='clusterfuzz-testcase-minimized-1.ext')
    self._test_download(
        self.minimized_key,
        testcase_id=self.testcase.key.id(),
        expect_filename='clusterfuzz-testcase-minimized-1.ext')
    # Only have access to minimized testcase.
    self._test_download(
        self.fuzzed_key,
        testcase_id=self.testcase.key.id(),
        expect_status=302,
        expect_blob=False)

    mock_issue.labels = issue_tracker.LabelStore(['restrict-view-commit'])
    self._test_download(
        testcase_id=self.testcase.key.id(),
        expect_status=302,
        expect_blob=False)
    self._test_download(
        self.minimized_key,
        testcase_id=self.testcase.key.id(),
        expect_status=302,
        expect_blob=False)
    self._test_download(
        self.fuzzed_key,
        testcase_id=self.testcase.key.id(),
        expect_status=302,
        expect_blob=False)

    # Privileged users should still be able to access everything.
    self.mock.get_user_email.return_value = 'a@user.com'
    self.mock.has_access.return_value = True

    self._test_download(
        testcase_id=self.testcase.key.id(),
        expect_filename='clusterfuzz-testcase-minimized-1.ext')
    self._test_download(
        self.minimized_key,
        testcase_id=self.testcase.key.id(),
        expect_filename='clusterfuzz-testcase-minimized-1.ext')
    self._test_download(
        self.fuzzed_key,
        testcase_id=self.testcase.key.id(),
        expect_filename='clusterfuzz-testcase-1.ext')

  def test_download_invalid_id(self):
    """Test download with an invalid testcase ID."""
    self.mock.has_access.return_value = True
    self._test_download(testcase_id=1337, expect_blob=False, expect_status=400)
    self._test_download(
        self.minimized_key,
        testcase_id=1337,
        expect_blob=False,
        expect_status=400)

  def test_download_mismatch_testcase_blob_key(self):
    """Test download with a blob key that doesn't match the testcase.."""
    self.mock.has_access.return_value = True
    self._test_download(
        self.unused_key,
        testcase_id=self.testcase.key.id(),
        expect_status=400,
        expect_blob=False)

  def test_download_filename(self):
    """Test filename of downloaded files."""
    self.mock.has_access.return_value = True

    expect_filename = 'file.ext'
    self._test_download(self.minimized_key, expect_filename=expect_filename)

    expect_filename = (
        'clusterfuzz-testcase-minimized-%s.ext' % self.testcase.key.id())
    self._test_download(
        testcase_id=self.testcase.key.id(), expect_filename=expect_filename)
    self._test_download(
        self.minimized_key,
        testcase_id=self.testcase.key.id(),
        expect_filename=expect_filename)

    expect_filename = ('clusterfuzz-testcase-%s.ext' % self.testcase.key.id())
    self._test_download(
        self.fuzzed_key,
        testcase_id=self.testcase.key.id(),
        expect_filename=expect_filename)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for train_rnn_generator_task."""

import os
import tempfile
import unittest

import pyfakefs.fake_filesystem_unittest as fake_fs_unittest

from clusterfuzz._internal.bot.fuzzers.ml.rnn import constants
from clusterfuzz._internal.bot.tasks import train_rnn_generator_task
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.system import shell
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils

MODEL_DIR = '/fake/model_directory'

# The directory containing sample corpus.
DATA_DIRECTORY = os.path.abspath(
    os.path.join(
        os.path.dirname(__file__), os.pardir, 'fuzzers', 'ml', 'rnn', 'data'))


class GetLastSavedModelTest(fake_fs_unittest.TestCase):
  """Get latest model test."""

  def setUp(self):
    """Setup for get last saved model test."""
    # Set up fake filesystem.
    test_utils.set_up_pyfakefs(self)

    # Create model directory.
    self.fs.create_dir(MODEL_DIR)

    # Create fake index file and data file.
    # Model_1 has two complete files.
    self.model_1_data_path = os.path.join(MODEL_DIR,
                                          'model_1.data-00000-of-00001')
    self.model_1_index_path = os.path.join(MODEL_DIR, 'model_1.index')

    # Create two files for model_1.
    self.fs.create_file(self.model_1_data_path)
    self.fs.create_file(self.model_1_index_path)

    # Update timestamp.
    os.utime(self.model_1_data_path, (1330711140, 1330711160))
    os.utime(self.model_1_index_path, (1330711140, 1330711160))

    # Model_2 has two complete files.
    self.model_2_data_path = os.path.join(MODEL_DIR,
                                          'model_2.data-00000-of-00001')
    self.model_2_index_path = os.path.join(MODEL_DIR, 'model_2.index')

    # Create two files for model_2.
    self.fs.create_file(self.model_2_data_path)
    self.fs.create_file(self.model_2_index_path)

    # Update timestamp. Make sure they are newer than model_1.
    os.utime(self.model_2_data_path, (1330713340, 1330713360))
    os.utime(self.model_2_index_path, (1330713340, 1330713360))

  def test_get_latest_model(self):
    """Test latest model is returned as a dictionary."""
    # Model_2 is newer than model_1, so we will get model_2.
    model_paths = train_rnn_generator_task.get_last_saved_model(MODEL_DIR)
    expected = {
        'data': self.model_2_data_path,
        'index': self.model_2_index_path
    }
    self.assertDictEqual(model_paths, expected)

  def test_get_valid_model(self):
    """Test lastest model is not returned if it is invalid."""
    # Remove one file from model_2, so model_2 is not valid.
    os.remove(self.model_2_index_path)
    self.assertFalse(os.path.exists(self.model_2_index_path))

    # Now we should get model_1.
    model_paths = train_rnn_generator_task.get_last_saved_model(MODEL_DIR)
    expected = {
        'data': self.model_1_data_path,
        'index': self.model_1_index_path
    }
    self.assertDictEqual(model_paths, expected)

  def test_no_valid_model(self):
    """Test no model is returned if all models are invalid."""
    # Remove one file from model_1 and one from model_2.
    os.remove(self.model_1_data_path)
    os.remove(self.model_2_index_path)
    self.assertFalse(os.path.exists(self.model_1_data_path))
    self.assertFalse(os.path.exists(self.model_2_index_path))

    # Now we should get empty dictionary since both models are invalid.
    model_paths = train_rnn_generator_task.get_last_saved_model(MODEL_DIR)
    expected = {}
    self.assertDictEqual(model_paths, expected)


@test_utils.with_cloud_emulators('datastore')
class ExecuteTaskTest(unittest.TestCase):
  """Execute training script test."""

  def setUp(self):
    test_helpers.patch_environ(self)

    self.fuzzer_name = 'fake_fuzzer'
    self.full_fuzzer_name = 'libFuzzer_fake_fuzzer'
    self.job_type = 'fake_job'
    self.temp_dir = tempfile.mkdtemp()

    data_types.FuzzTarget(
        engine='libFuzzer', binary='fake_fuzzer', project='test-project').put()

    os.environ['FUZZ_INPUTS_DISK'] = self.temp_dir

    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.tasks.ml_train_utils.get_corpus',
        'clusterfuzz._internal.bot.tasks.train_rnn_generator_task.train_rnn',
        'clusterfuzz._internal.bot.tasks.train_rnn_generator_task.upload_model_to_gcs',
    ])

    self.mock.get_corpus.return_value = True
    self.mock.train_rnn.return_value = new_process.ProcessResult(return_code=0)
    self.mock.upload_model_to_gcs.return_value = True

  def tearDown(self):
    shell.remove_directory(self.temp_dir)

  def test_execute(self):
    """Test execute task."""
    input_directory = os.path.join(
        self.temp_dir,
        self.fuzzer_name + train_rnn_generator_task.CORPUS_DIR_SUFFIX)
    model_directory = os.path.join(
        self.temp_dir,
        self.fuzzer_name + train_rnn_generator_task.MODEL_DIR_SUFFIX)
    log_directory = os.path.join(
        self.temp_dir,
        self.fuzzer_name + train_rnn_generator_task.LOG_DIR_SUFFIX)

    train_rnn_generator_task.execute_task(self.full_fuzzer_name, self.job_type)

    self.mock.train_rnn.assert_called_once_with(input_directory,
                                                model_directory, log_directory)


@unittest.skipIf(not os.getenv('ML_TESTS'), 'ML_TESTS=1 must be set')
@test_utils.integration
class MLRnnTrainTaskIntegrationTest(unittest.TestCase):
  """ML RNN training integration tests."""

  def setUp(self):
    self.input_directory = os.path.join(DATA_DIRECTORY, 'input')
    self.model_directory = tempfile.mkdtemp()
    self.log_directory = tempfile.mkdtemp()

    self.batch_size = 1
    self.hidden_state_size = 2
    self.hidden_layer_size = 1

  def tearDown(self):
    shell.remove_directory(self.model_directory)
    shell.remove_directory(self.log_directory)

  def test_train_rnn(self):
    """Test train RNN model on a simple corpus."""
    # No model exists in model directory.
    self.assertFalse(
        train_rnn_generator_task.get_last_saved_model(self.model_directory))

    # The training should be fast (a few seconds) since sample corpus is
    # extremely small.
    result = train_rnn_generator_task.train_rnn(
        self.input_directory, self.model_directory, self.log_directory,
        self.batch_size, self.hidden_state_size, self.hidden_layer_size)

    self.assertEqual(result.return_code, constants.ExitCode.SUCCESS)
    self.assertFalse(result.timed_out)

    # At least one model exists.
    self.assertTrue(
        train_rnn_generator_task.get_last_saved_model(self.model_directory))

  def test_small_corpus(self):
    """Test small corpus situation."""
    # Increase batch size so the sample corpus appears small in this case.
    self.batch_size = 100

    result = train_rnn_generator_task.train_rnn(
        self.input_directory, self.model_directory, self.log_directory,
        self.batch_size, self.hidden_state_size, self.hidden_layer_size)

    self.assertEqual(result.return_code, constants.ExitCode.CORPUS_TOO_SMALL)
    self.assertFalse(result.timed_out)

    # No model exsits after execution.
    self.assertFalse(
        train_rnn_generator_task.get_last_saved_model(self.model_directory))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for libfuzzer."""

import mock
from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.bot.fuzzers import options
from clusterfuzz._internal.tests.test_libs import test_utils


class FuzzerOptionsTest(fake_filesystem_unittest.TestCase):
  """FuzzerOptions tests."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)

  @mock.patch('random.SystemRandom')
  def test_basic(self, mock_sysrandom):
    """Basic test."""
    mock_sysrandom.return_value.randint = lambda x, y: 1337

    input_data = ('[libfuzzer]\n'
                  'max_len=9001\n'
                  'dict=blah.dict\n'
                  'blah=9002\n'
                  'rand=random(1, 100)\n'
                  '[asan]\n'
                  'detect_leaks=0\n'
                  '[msan]\n'
                  'msan_option=1\n'
                  '[ubsan]\n'
                  'ubsan_option=0\n')

    self.fs.create_file('/path/blah.options', contents=input_data)
    self.fs.create_file('/path/blah.dict', contents=input_data)
    fuzzer_options = options.FuzzerOptions('/path/blah.options')

    fuzzer_arguments = fuzzer_options.get_engine_arguments('libfuzzer')

    self.assertListEqual(
        sorted(fuzzer_arguments.list()),
        ['-blah=9002', '-dict=/path/blah.dict', '-max_len=9001', '-rand=1337'])
    self.assertDictEqual(
        fuzzer_arguments.dict(), {
            'blah': '9002',
            'dict': '/path/blah.dict',
            'max_len': '9001',
            'rand': '1337'
        })
    self.assertDictEqual(fuzzer_options.get_asan_options(),
                         {'detect_leaks': '0'})
    self.assertDictEqual(fuzzer_options.get_msan_options(),
                         {'msan_option': '1'})
    self.assertDictEqual(fuzzer_options.get_ubsan_options(),
                         {'ubsan_option': '0'})
    self.assertEqual(fuzzer_arguments['max_len'], '9001')
    self.assertEqual(fuzzer_arguments.get('max_len', constructor=int), 9001)
    self.assertEqual(fuzzer_arguments.get('noexist', constructor=int), None)


class GetFuzzTargetOptions(fake_filesystem_unittest.TestCase):
  """get_fuzz_target_options tests."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)

    input_data = ('[libfuzzer]\nclose_fd_mask=1\n')
    self.fs.create_file('/path/fuzz_target.options', contents=input_data)

  def _get_arguments(self, fuzz_target_path):
    """Helper to return fuzz target arguments by parsing options file for a
    fuzz target."""
    fuzzer_options = options.get_fuzz_target_options(fuzz_target_path)
    if fuzzer_options is None:
      return None

    fuzzer_arguments = fuzzer_options.get_engine_arguments('libfuzzer')
    return sorted(fuzzer_arguments.list())

  def test_without_extension(self):
    self.assertEqual(
        self._get_arguments('/path/fuzz_target'), ['-close_fd_mask=1'])

  def test_with_extension(self):
    self.assertEqual(
        self._get_arguments('/path/fuzz_target.exe'), ['-close_fd_mask=1'])

  def test_not_exist(self):
    self.assertEqual(self._get_arguments('/path/not_exist'), None)
    self.assertEqual(self._get_arguments('/path/not_exist.exe'), None)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzer log utilities."""

import datetime

from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

DATE_FORMAT = '%Y-%m-%d'
TIME_FORMAT = '%H:%M:%S:%f'

LOG_EXTENSION = '.log'
LOG_PATH_FORMAT = DATE_FORMAT + '/' + TIME_FORMAT


def get_bucket():
  """Return path to fuzzer logs bucket."""
  return local_config.ProjectConfig().get('logs.fuzzer.bucket')


def get_log_relative_path(log_time, file_extension=None):
  """Generate a relative path for a log using the given time.
  Args:
    log_time: A datetime object.
    file_extension: A string appended to the end of the log file name.
      LOG_EXTENSION is used if None.
  Returns:
    A string containing name of the log file.
  """
  if file_extension is None:
    file_extension = LOG_EXTENSION

  return log_time.strftime(LOG_PATH_FORMAT) + file_extension


def get_logs_directory(bucket_name, fuzzer_name, job_type=None, logs_date=None):
  """Get directory path of logs for a given fuzzer/job.
  Args:
    bucket_name: Bucket logs are stored in.
    fuzzer_name: Name of the fuzzer.
    job_type: Job name.
    logs_date: Optional datetime.date for the logs.

  Returns:
    A cloud storage path to the directory containing the desired logs. Format
    returned is /{bucket name}/{path}.
  """
  path = '/%s/%s' % (bucket_name, fuzzer_name)
  if job_type:
    path += '/%s' % job_type

  if logs_date is not None:
    assert job_type is not None
    path += '/%s' % logs_date

  return path


def upload_to_logs(bucket_name,
                   contents,
                   time=None,
                   fuzzer_name=None,
                   job_type=None,
                   file_extension=None):
  """Upload file contents to log directory in GCS bucket.
  Args:
    bucket_name: Bucket logs are stored in.
    contents: String containing log to be uploaded.
    time: A datetime object used to generate filename for the log.
    fuzzer_name: Name of the fuzzer. If None, gets this from the environment.
    job_type: Job name. If None, gets this from the environment.
    file_extension: A string appended to the end of the log filename. A default
      value is used if None.

  Returns:
    The path of the uploaded file and whether the uploaded succeeded.
  """
  if not fuzzer_name:
    fuzzer_name = environment.get_value('FUZZER_NAME')

  if not job_type:
    job_type = environment.get_value('JOB_NAME')

  log_directory = get_logs_directory(bucket_name, fuzzer_name, job_type)

  if not time:
    time = datetime.datetime.utcnow()

  log_path = 'gs:/' + log_directory + '/' + get_log_relative_path(
      time, file_extension)

  if storage.write_data(contents, log_path):
    logs.log('Uploaded file to logs bucket.', log_path=log_path)
  else:
    logs.log_error('Failed to write file to logs bucket.', log_path=log_path)


def upload_script_log(log_contents, fuzzer_name=None, job_type=None):
  """Upload logs to script logs GCS bucket.
  Args:
    log_contents: String containing log to be uploaded.
    fuzzer_name: Name of the fuzzer. If None, gets this from the environment.
    job_type: Job name. If None, gets this from the environment.
  """
  logs_bucket = get_bucket()
  if logs_bucket and log_contents and log_contents.strip():
    upload_to_logs(
        logs_bucket, log_contents, fuzzer_name=fuzzer_name, job_type=job_type)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sanitizer related functions."""

import os

from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

from . import adb
from . import constants
from . import settings

try:
  from clusterfuzz._internal.system import new_process
except ImportError:
  # On App Engine.
  new_process = None

ASAN_SCRIPT_TIMEOUT = 15 * 60
SANITIZER_TOOL_TO_FILE_MAPPINGS = {
    'asan': 'asan.options',
}


def get_ld_library_path_for_sanitizers():
  """Return LD_LIBRARY_PATH setting for sanitizers, None otherwise."""
  if not settings.get_sanitizer_tool_name():
    return None

  sanitizer_libs = adb.run_shell_command(
      ['find', '/system/lib64/', '-name', '*san_standalone*'])
  if not sanitizer_libs:
    return None

  sanitizer_lib_directory = os.path.dirname(sanitizer_libs.splitlines()[0])
  return '/system/lib64:' + sanitizer_lib_directory


def get_options_file_path(sanitizer_tool_name):
  """Return path for the sanitizer options file."""
  # If this a full sanitizer system build, then update the options file in
  # /system, else just put it in device temp directory.
  sanitizer_directory = ('/system' if settings.get_sanitizer_tool_name() else
                         constants.DEVICE_TMP_DIR)

  sanitizer_filename = SANITIZER_TOOL_TO_FILE_MAPPINGS.get(
      sanitizer_tool_name.lower())
  if sanitizer_filename is None:
    logs.log_error('Unsupported sanitizer: ' + sanitizer_tool_name)
    return None

  return os.path.join(sanitizer_directory, sanitizer_filename)


def set_options(sanitizer_tool_name, sanitizer_options):
  """Set sanitizer options on the disk file."""
  sanitizer_options_file_path = get_options_file_path(sanitizer_tool_name)
  if not sanitizer_options_file_path:
    return

  adb.write_data_to_file(sanitizer_options, sanitizer_options_file_path)


def setup_asan_if_needed():
  """Set up asan on device."""
  if not environment.get_value('ASAN_DEVICE_SETUP'):
    # Only do this step if explicitly enabled in the job type. This cannot be
    # determined from libraries in application directory since they can go
    # missing in a bad build, so we want to catch that.
    return

  if settings.get_sanitizer_tool_name():
    # If this is a sanitizer build, no need to setup ASAN (incompatible).
    return

  app_directory = environment.get_value('APP_DIR')
  if not app_directory:
    # No app directory -> No ASAN runtime library. No work to do, bail out.
    return

  # Initialize variables.
  android_directory = environment.get_platform_resources_directory()
  device_id = environment.get_value('ANDROID_SERIAL')

  # Execute the script.
  logs.log('Executing ASan device setup script.')
  asan_device_setup_script_path = os.path.join(android_directory, 'third_party',
                                               'asan_device_setup.sh')
  extra_options_arg = 'include_if_exists=' + get_options_file_path('asan')
  asan_device_setup_script_args = [
      '--lib', app_directory, '--device', device_id, '--extra-options',
      extra_options_arg
  ]

  process = new_process.ProcessRunner(asan_device_setup_script_path,
                                      asan_device_setup_script_args)
  result = process.run_and_wait()
  if result.return_code:
    logs.log_error('Failed to setup ASan on device.', output=result.output)
    return

  logs.log(
      'ASan device setup script successfully finished, waiting for boot.',
      output=result.output)

  # Wait until fully booted as otherwise shell restart followed by a quick
  # reboot can trigger data corruption in /data/data.
  adb.wait_until_fully_booted()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unpack task for unpacking a multi-testcase archive
from user upload."""

import json
import os

from google.cloud import ndb

from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import blobs
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import archive
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell


def execute_task(metadata_id, job_type):
  """Unpack a bundled testcase archive and create analyze jobs for each item."""
  metadata = ndb.Key(data_types.BundledArchiveMetadata, int(metadata_id)).get()
  if not metadata:
    logs.log_error('Invalid bundle metadata id %s.' % metadata_id)
    return

  bot_name = environment.get_value('BOT_NAME')
  upload_metadata = data_types.TestcaseUploadMetadata.query(
      data_types.TestcaseUploadMetadata.blobstore_key ==
      metadata.blobstore_key).get()
  if not upload_metadata:
    logs.log_error('Invalid upload metadata key %s.' % metadata.blobstore_key)
    return

  job = data_types.Job.query(data_types.Job.name == metadata.job_type).get()
  if not job:
    logs.log_error('Invalid job_type %s.' % metadata.job_type)
    return

  # Update the upload metadata with this bot name.
  upload_metadata.bot_name = bot_name
  upload_metadata.put()

  # We can't use FUZZ_INPUTS directory since it is constrained
  # by tmpfs limits.
  testcases_directory = environment.get_value('FUZZ_INPUTS_DISK')

  # Retrieve multi-testcase archive.
  archive_path = os.path.join(testcases_directory, metadata.archive_filename)
  if not blobs.read_blob_to_disk(metadata.blobstore_key, archive_path):
    logs.log_error('Could not retrieve archive for bundle %d.' % metadata_id)
    tasks.add_task('unpack', metadata_id, job_type)
    return

  try:
    archive.unpack(archive_path, testcases_directory)
  except:
    logs.log_error('Could not unpack archive for bundle %d.' % metadata_id)
    tasks.add_task('unpack', metadata_id, job_type)
    return

  # Get additional testcase metadata (if any).
  additional_metadata = None
  if upload_metadata.additional_metadata_string:
    additional_metadata = json.loads(upload_metadata.additional_metadata_string)

  archive_state = data_types.ArchiveStatus.NONE
  bundled = True
  file_list = archive.get_file_list(archive_path)

  for file_path in file_list:
    absolute_file_path = os.path.join(testcases_directory, file_path)
    filename = os.path.basename(absolute_file_path)

    # Only files are actual testcases. Skip directories.
    if not os.path.isfile(absolute_file_path):
      continue

    try:
      file_handle = open(absolute_file_path, 'rb')
      blob_key = blobs.write_blob(file_handle)
      file_handle.close()
    except:
      blob_key = None

    if not blob_key:
      logs.log_error(
          'Could not write testcase %s to blobstore.' % absolute_file_path)
      continue

    data_handler.create_user_uploaded_testcase(
        blob_key, metadata.blobstore_key, archive_state,
        metadata.archive_filename, filename, metadata.timeout, job,
        metadata.job_queue, metadata.http_flag, metadata.gestures,
        metadata.additional_arguments, metadata.bug_information,
        metadata.crash_revision, metadata.uploader_email, metadata.platform_id,
        metadata.app_launch_command, metadata.fuzzer_name,
        metadata.overridden_fuzzer_name, metadata.fuzzer_binary_name, bundled,
        upload_metadata.retries, upload_metadata.bug_summary_update_flag,
        upload_metadata.quiet_flag, additional_metadata)

  # The upload metadata for the archive is not needed anymore since we created
  # one for each testcase.
  upload_metadata.key.delete()

  shell.clear_testcase_directories()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common functions for task creation for test cases."""

from clusterfuzz._internal.base import bisection
from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment


def mark_unreproducible_if_flaky(testcase, potentially_flaky):
  """Check to see if a test case appears to be flaky."""
  task_name = environment.get_value('TASK_NAME')

  # If this run does not suggest that we are flaky, clear the flag and assume
  # that we are reproducible.
  if not potentially_flaky:
    testcase.set_metadata('potentially_flaky', False)
    return

  # If we have not been marked as potentially flaky in the past, don't mark
  # mark the test case as unreproducible yet. It is now potentially flaky.
  if not testcase.get_metadata('potentially_flaky'):
    testcase.set_metadata('potentially_flaky', True)

    # In this case, the current task will usually be in a state where it cannot
    # be completed. Recreate it.
    tasks.add_task(task_name, testcase.key.id(), testcase.job_type)
    return

  # At this point, this test case has been flagged as potentially flaky twice.
  # It should be marked as unreproducible. Mark it as unreproducible, and set
  # fields that cannot be populated accordingly.
  if task_name == 'minimize' and not testcase.minimized_keys:
    testcase.minimized_keys = 'NA'
  if task_name in ['minimize', 'impact']:
    testcase.set_impacts_as_na()
  if task_name in ['minimize', 'regression']:
    testcase.regression = 'NA'
  if task_name in ['minimize', 'progression']:
    testcase.fixed = 'NA'

  testcase.one_time_crasher_flag = True
  data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                       'Testcase appears to be flaky')

  # Issue update to flip reproducibility label is done in App Engine cleanup
  # cron. This avoids calling the issue tracker apis from GCE.

  # For unreproducible testcases, it is still beneficial to get component
  # information from blame task.
  create_blame_task_if_needed(testcase)

  # Let bisection service know about flakiness.
  bisection.request_bisection(testcase)


def create_blame_task_if_needed(testcase):
  """Creates a blame task if needed."""
  # Blame doesn't work for non-chromium projects.
  if not utils.is_chromium():
    return

  # Blame is only applicable to chromium project, otherwise bail out.
  if testcase.project_name != 'chromium':
    return

  # We cannot run blame job for custom binaries since we don't have any context
  # on the crash revision and regression range.
  if build_manager.is_custom_binary():
    return

  # Don't send duplicate issues to Predator. This causes issues with metrics
  # tracking and wastes cycles.
  if testcase.status == 'Duplicate':
    return

  create_task = False
  if testcase.one_time_crasher_flag:
    # For unreproducible testcases, it is still beneficial to get component
    # information from blame task.
    create_task = True
  else:
    # Reproducible testcase.
    # Step 1: Check if the regression task finished. If not, bail out.
    if not testcase.regression:
      return

    # Step 2: Check if the symbolize task is applicable and finished. If not,
    # bail out.
    if build_manager.has_symbolized_builds() and not testcase.symbolized:
      return

    create_task = True

  if create_task:
    tasks.add_task('blame', testcase.key.id(), testcase.job_type)


def create_impact_task_if_needed(testcase):
  """Creates an impact task if needed."""
  # Impact doesn't make sense for non-chromium projects.
  if not utils.is_chromium():
    return

  # Impact is only applicable to chromium project, otherwise bail out.
  if testcase.project_name != 'chromium':
    return

  # We cannot run impact job for custom binaries since we don't have any
  # archived production builds for these.
  if build_manager.is_custom_binary():
    return

  tasks.add_task('impact', testcase.key.id(), testcase.job_type)


def create_minimize_task_if_needed(testcase):
  """Creates a minimize task if needed."""
  tasks.add_task('minimize', testcase.key.id(), testcase.job_type)


def create_regression_task_if_needed(testcase):
  """Creates a regression task if needed."""
  # We cannot run regression job for custom binaries since we don't have any
  # archived builds for previous revisions. We only track the last uploaded
  # custom build.
  if build_manager.is_custom_binary():
    return

  tasks.add_task('regression', testcase.key.id(), testcase.job_type)


def create_variant_tasks_if_needed(testcase):
  """Creates a variant task if needed."""
  if testcase.duplicate_of:
    # If another testcase exists with same params, no need to spend cycles on
    # calculating variants again.
    return

  testcase_id = testcase.key.id()
  project = data_handler.get_project_name(testcase.job_type)
  jobs = data_types.Job.query(data_types.Job.project == project)
  for job in jobs:
    # The variant needs to be tested in a different job type than us.
    job_type = job.name
    if testcase.job_type == job_type:
      continue

    # Don't try to reproduce engine fuzzer testcase with blackbox fuzzer
    # testcases and vice versa.
    if (environment.is_engine_fuzzer_job(testcase.job_type) !=
        environment.is_engine_fuzzer_job(job_type)):
      continue

    # Skip experimental jobs.
    job_environment = job.get_environment()
    if utils.string_is_true(job_environment.get('EXPERIMENTAL')):
      continue

    queue = tasks.queue_for_platform(job.platform)
    tasks.add_task('variant', testcase_id, job_type, queue)

    variant = data_handler.get_testcase_variant(testcase_id, job_type)
    variant.status = data_types.TestcaseVariantStatus.PENDING
    variant.put()


def create_symbolize_task_if_needed(testcase):
  """Creates a symbolize task if needed."""
  # We cannot run symbolize job for custom binaries since we don't have any
  # archived symbolized builds.
  if build_manager.is_custom_binary():
    return

  # Make sure we have atleast one symbolized url pattern defined in job type.
  if not build_manager.has_symbolized_builds():
    return

  tasks.add_task('symbolize', testcase.key.id(), testcase.job_type)


def create_tasks(testcase):
  """Create tasks like minimization, regression, impact, progression, stack
  stack for a newly generated testcase."""
  # No need to create progression task. It is automatically created by the cron
  # handler for reproducible testcases.

  # For a non reproducible crash.
  if testcase.one_time_crasher_flag:
    # For unreproducible testcases, it is still beneficial to get component
    # information from blame task.
    create_blame_task_if_needed(testcase)
    return

  # For a fully reproducible crash.

  # MIN environment variable defined in a job definition indicates if
  # we want to do the heavy weight tasks like minimization, regression,
  # impact, etc on this testcase. These are usually skipped when we have
  # a large timeout and we can't afford to waste more than a couple of hours
  # on these jobs.
  testcase_id = testcase.key.id()
  if environment.get_value('MIN') == 'No':
    testcase = data_handler.get_testcase_by_id(testcase_id)
    testcase.minimized_keys = 'NA'
    testcase.regression = 'NA'
    testcase.set_impacts_as_na()
    testcase.put()
    return

  # Just create the minimize task for now. Once minimization is complete, it
  # automatically created the rest of the needed tasks.
  create_minimize_task_if_needed(testcase)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""issue_filer tests."""

# pylint: disable=protected-access

import datetime
import os
import unittest

import mock
import parameterized
import six

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import pubsub
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import mock_config
from clusterfuzz._internal.tests.test_libs import test_utils
from libs.issue_management import issue_filer
from libs.issue_management import issue_tracker_policy
from libs.issue_management import monorail
from libs.issue_management.issue_tracker import LabelStore
from libs.issue_management.monorail.issue import Issue as MonorailIssue

CHROMIUM_POLICY = issue_tracker_policy.IssueTrackerPolicy({
    'status': {
        'assigned': 'Assigned',
        'duplicate': 'Duplicate',
        'verified': 'Verified',
        'new': 'Untriaged',
        'wontfix': 'WontFix',
        'fixed': 'Fixed'
    },
    'all': {
        'status': 'new',
        'labels': ['ClusterFuzz', 'Stability-%SANITIZER%']
    },
    'non_security': {
        'labels': ['Type-Bug'],
        'crash_labels': ['Stability-Crash', 'Pri-1'],
        'non_crash_labels': ['Pri-2']
    },
    'labels': {
        'ignore': 'ClusterFuzz-Ignore',
        'verified': 'ClusterFuzz-Verified',
        'security_severity': 'Security_Severity-%SEVERITY%',
        'needs_feedback': 'Needs-Feedback',
        'invalid_fuzzer': 'ClusterFuzz-Invalid-Fuzzer',
        'reported': None,
        'wrong': 'ClusterFuzz-Wrong',
        'fuzz_blocker': 'Fuzz-Blocker',
        'reproducible': 'Reproducible',
        'auto_cc_from_owners': 'ClusterFuzz-Auto-CC',
        'os': 'OS-%PLATFORM%',
        'unreproducible': 'Unreproducible',
        'restrict_view': 'Restrict-View-SecurityTeam'
    },
    'security': {
        'labels': ['Type-Bug-Security']
    },
    'existing': {
        'labels': ['Stability-%SANITIZER%']
    },
    'unreproducible_component': 'Unreproducible>Component'
})

OSS_FUZZ_POLICY = issue_tracker_policy.IssueTrackerPolicy({
    'status': {
        'assigned': 'Assigned',
        'duplicate': 'Duplicate',
        'verified': 'Verified',
        'new': 'New',
        'wontfix': 'WontFix',
        'fixed': 'Fixed'
    },
    'all': {
        'status': 'new',
        'labels': ['ClusterFuzz', 'Stability-%SANITIZER%'],
        'issue_body_footer':
            'When you fix this bug, please\n'
            '  * mention the fix revision(s).\n'
            '  * state whether the bug was a short-lived regression or an old '
            'bug in any stable releases.\n'
            '  * add any other useful information.\n'
            'This information can help downstream consumers.\n\n'
            'If you need to contact the OSS-Fuzz team with a question, '
            'concern, or any other feedback, please file an issue at '
            'https://github.com/google/oss-fuzz/issues.'
    },
    'non_security': {
        'labels': ['Type-Bug']
    },
    'labels': {
        'ignore': 'ClusterFuzz-Ignore',
        'verified': 'ClusterFuzz-Verified',
        'security_severity': 'Security_Severity-%SEVERITY%',
        'needs_feedback': 'Needs-Feedback',
        'invalid_fuzzer': 'ClusterFuzz-Invalid-Fuzzer',
        'reported': 'Reported-%YYYY-MM-DD%',
        'wrong': 'ClusterFuzz-Wrong',
        'fuzz_blocker': 'Fuzz-Blocker',
        'reproducible': 'Reproducible',
        'auto_cc_from_owners': 'ClusterFuzz-Auto-CC',
        'os': 'OS-%PLATFORM%',
        'unreproducible': 'Unreproducible',
        'restrict_view': 'Restrict-View-Commit'
    },
    'security': {
        'labels': ['Type-Bug-Security']
    },
    'deadline_policy_message':
        'This bug is subject to a 90 day disclosure deadline. If 90 days '
        'elapse\n'
        'without an upstream patch, then the bug report will automatically\n'
        'become visible to the public.',
    'existing': {
        'labels': ['Stability-%SANITIZER%']
    }
})

DEADLINE_NOTE = (
    'This bug is subject to a 90 day disclosure deadline. If 90 days elapse\n'
    'without an upstream patch, then the bug report will automatically\n'
    'become visible to the public.')

FIX_NOTE = (
    'When you fix this bug, please\n'
    '  * mention the fix revision(s).\n'
    '  * state whether the bug was a short-lived regression or an old bug'
    ' in any stable releases.\n'
    '  * add any other useful information.\n'
    'This information can help downstream consumers.')

QUESTIONS_NOTE = (
    'If you need to contact the OSS-Fuzz team with a question, concern, or any '
    'other feedback, please file an issue at '
    'https://github.com/google/oss-fuzz/issues.')


class IssueTrackerManager(object):
  """Mock issue tracker manager."""

  def __init__(self, project_name):
    self.project_name = project_name
    self.last_issue = None

  def save(self, issue, *args, **kwargs):  # pylint: disable=unused-argument
    self.last_issue = issue


@test_utils.with_cloud_emulators('datastore')
class IssueFilerTests(unittest.TestCase):
  """Tests for the issue filer."""

  def setUp(self):
    data_types.Job(
        name='job1',
        environment_string='ISSUE_VIEW_RESTRICTIONS = all\nPROJECT_NAME = proj',
        platform='linux').put()

    data_types.Job(
        name='job2',
        environment_string='ISSUE_VIEW_RESTRICTIONS = security',
        platform='linux').put()

    data_types.Job(
        name='job3',
        environment_string='ISSUE_VIEW_RESTRICTIONS = none',
        platform='linux').put()

    data_types.Job(
        name='chromeos_job4', environment_string='', platform='linux').put()

    testcase_args = {
        'crash_type': 'Heap-use-after-free',
        'crash_address': '0x1337',
        'crash_state': '1\n2\n3\n',
        'crash_stacktrace': 'stack\n',
        'fuzzer_name': 'fuzzer',
    }

    self.testcase1 = data_types.Testcase(job_type='job1', **testcase_args)
    self.testcase1.put()

    self.testcase1_security = data_types.Testcase(
        security_flag=True, job_type='job1', **testcase_args)
    self.testcase1_security.put()

    self.testcase2 = data_types.Testcase(job_type='job2', **testcase_args)
    self.testcase2.put()

    self.testcase2_security = data_types.Testcase(
        security_flag=True, job_type='job2', **testcase_args)
    self.testcase2_security.put()

    self.testcase3 = data_types.Testcase(job_type='job3', **testcase_args)
    self.testcase3.put()

    self.testcase3_security = data_types.Testcase(
        job_type='job3', security_flag=True, **testcase_args)
    self.testcase3_security.put()

    self.testcase4 = data_types.Testcase(
        job_type='chromeos_job4', **testcase_args)
    self.testcase4.put()

    self.testcase5 = data_types.Testcase(
        job_type='job',
        additional_metadata=('{"issue_labels": "label1 , label2,,", '
                             '"issue_components": "component1,component2"}'),
        **testcase_args)
    self.testcase5.put()

    self.testcase6 = data_types.Testcase(
        job_type='job', additional_metadata='invalid', **testcase_args)
    self.testcase6.put()

    data_types.ExternalUserPermission(
        email='user@example.com',
        entity_name='job2',
        entity_kind=data_types.PermissionEntityKind.JOB,
        is_prefix=False,
        auto_cc=data_types.AutoCCType.ALL).put()

    data_types.ExternalUserPermission(
        email='user2@example.com',
        entity_name='job3',
        entity_kind=data_types.PermissionEntityKind.JOB,
        is_prefix=False,
        auto_cc=data_types.AutoCCType.SECURITY).put()

    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.utcnow',
        'clusterfuzz._internal.datastore.data_handler.get_issue_description',
        'libs.issue_management.issue_tracker_policy.get',
    ])

    self.mock.get_issue_description.return_value = 'Issue'
    self.mock.utcnow.return_value = datetime.datetime(2016, 1, 1)

  def test_filed_issues_chromium(self):
    """Tests issue filing for chromium."""
    self.mock.get.return_value = CHROMIUM_POLICY
    issue_tracker = monorail.IssueTracker(IssueTrackerManager('chromium'))
    issue_filer.file_issue(self.testcase4, issue_tracker)
    self.assertIn('OS-Chrome', issue_tracker._itm.last_issue.labels)
    self.assertEqual('Untriaged', issue_tracker._itm.last_issue.status)
    self.assertNotIn('Restrict-View-SecurityTeam',
                     issue_tracker._itm.last_issue.labels)

  def test_filed_issues_chromium_security(self):
    """Tests issue filing for chromium."""
    self.testcase4.security_flag = True
    self.testcase4.put()
    self.mock.get.return_value = CHROMIUM_POLICY
    issue_tracker = monorail.IssueTracker(IssueTrackerManager('chromium'))
    issue_filer.file_issue(self.testcase4, issue_tracker)
    self.assertIn('OS-Chrome', issue_tracker._itm.last_issue.labels)
    self.assertEqual('Untriaged', issue_tracker._itm.last_issue.status)
    self.assertIn('Restrict-View-SecurityTeam',
                  issue_tracker._itm.last_issue.labels)

  def test_filed_issues_oss_fuzz(self):
    """Tests issue filing for oss-fuzz."""
    self.mock.get.return_value = OSS_FUZZ_POLICY
    issue_tracker = monorail.IssueTracker(IssueTrackerManager('oss-fuzz'))
    issue_filer.file_issue(self.testcase1, issue_tracker)
    self.assertEqual('New', issue_tracker._itm.last_issue.status)
    self.assertTrue(
        issue_tracker._itm.last_issue.has_label_matching(
            'restrict-view-commit'))
    self.assertFalse(
        issue_tracker._itm.last_issue.has_label_matching('reported-2016-01-01'))
    self.assertNotIn(DEADLINE_NOTE, issue_tracker._itm.last_issue.body)
    self.assertIn(FIX_NOTE, issue_tracker._itm.last_issue.body)
    self.assertIn(QUESTIONS_NOTE, issue_tracker._itm.last_issue.body)

    issue_filer.file_issue(self.testcase1_security, issue_tracker)
    self.assertTrue(
        issue_tracker._itm.last_issue.has_label_matching(
            'restrict-view-commit'))
    self.assertFalse(
        issue_tracker._itm.last_issue.has_label_matching('reported-2016-01-01'))
    self.assertNotIn(DEADLINE_NOTE, issue_tracker._itm.last_issue.body)
    self.assertIn(FIX_NOTE, issue_tracker._itm.last_issue.body)
    self.assertIn(QUESTIONS_NOTE, issue_tracker._itm.last_issue.body)

    issue_filer.file_issue(self.testcase2, issue_tracker)
    self.assertFalse(
        issue_tracker._itm.last_issue.has_label_matching(
            'restrict-view-commit'))
    self.assertTrue(
        issue_tracker._itm.last_issue.has_label_matching('reported-2016-01-01'))
    self.assertNotIn(DEADLINE_NOTE, issue_tracker._itm.last_issue.body)
    self.assertIn(FIX_NOTE, issue_tracker._itm.last_issue.body)
    self.assertIn(QUESTIONS_NOTE, issue_tracker._itm.last_issue.body)

    issue_filer.file_issue(self.testcase2_security, issue_tracker)
    self.assertTrue(
        issue_tracker._itm.last_issue.has_label_matching(
            'restrict-view-commit'))
    self.assertTrue(
        issue_tracker._itm.last_issue.has_label_matching('reported-2016-01-01'))
    self.assertIn(DEADLINE_NOTE, issue_tracker._itm.last_issue.body)
    self.assertIn(FIX_NOTE, issue_tracker._itm.last_issue.body)
    self.assertIn(QUESTIONS_NOTE, issue_tracker._itm.last_issue.body)

    issue_filer.file_issue(self.testcase3, issue_tracker)
    self.assertFalse(
        issue_tracker._itm.last_issue.has_label_matching(
            'restrict-view-commit'))
    self.assertFalse(
        issue_tracker._itm.last_issue.has_label_matching('reported-2016-01-01'))
    self.assertNotIn(DEADLINE_NOTE, issue_tracker._itm.last_issue.body)
    self.assertIn(FIX_NOTE, issue_tracker._itm.last_issue.body)
    self.assertIn(QUESTIONS_NOTE, issue_tracker._itm.last_issue.body)

    issue_filer.file_issue(self.testcase3_security, issue_tracker)
    self.assertFalse(
        issue_tracker._itm.last_issue.has_label_matching(
            'restrict-view-commit'))
    self.assertTrue(
        issue_tracker._itm.last_issue.has_label_matching('reported-2016-01-01'))
    self.assertNotIn(DEADLINE_NOTE, issue_tracker._itm.last_issue.body)
    self.assertIn(FIX_NOTE, issue_tracker._itm.last_issue.body)
    self.assertIn(QUESTIONS_NOTE, issue_tracker._itm.last_issue.body)

  def test_testcase_metadata_labels_and_components(self):
    """Tests issue filing with additional labels and components."""
    self.mock.get.return_value = CHROMIUM_POLICY
    issue_tracker = monorail.IssueTracker(IssueTrackerManager('chromium'))
    issue_filer.file_issue(self.testcase5, issue_tracker)
    six.assertCountEqual(self, [
        'ClusterFuzz',
        'Reproducible',
        'Pri-1',
        'Stability-Crash',
        'Type-Bug',
        'label1',
        'label2',
    ], issue_tracker._itm.last_issue.labels)
    six.assertCountEqual(self, [
        'component1',
        'component2',
    ], issue_tracker._itm.last_issue.components)

  def test_testcase_metadata_invalid(self):
    """Tests issue filing with invalid metadata."""
    self.mock.get.return_value = CHROMIUM_POLICY
    issue_tracker = monorail.IssueTracker(IssueTrackerManager('chromium'))
    issue_filer.file_issue(self.testcase6, issue_tracker)
    six.assertCountEqual(
        self,
        ['ClusterFuzz', 'Reproducible', 'Pri-1', 'Stability-Crash', 'Type-Bug'],
        issue_tracker._itm.last_issue.labels)

  @parameterized.parameterized.expand([
      ('chromium', CHROMIUM_POLICY),
      ('oss-fuzz', OSS_FUZZ_POLICY),
  ])
  def test_security_severity_functional_bug(self, project_name, policy):
    """Test security severity label is not set for a functional bug."""
    self.mock.get.return_value = policy
    issue_tracker = monorail.IssueTracker(IssueTrackerManager(project_name))

    self.testcase1.security_flag = False
    self.testcase1.security_severity = None
    self.testcase1.put()
    issue_filer.file_issue(self.testcase1, issue_tracker)
    self.assertFalse(
        issue_tracker._itm.last_issue.has_label_by_prefix('Security_Severity-'))

  @parameterized.parameterized.expand([
      ('chromium', CHROMIUM_POLICY),
      ('oss-fuzz', OSS_FUZZ_POLICY),
  ])
  def test_security_severity_security_bug_default_severity(
      self, project_name, policy):
    """Test security severity label is set when testcase is a security bug and
    no severity can be determined."""
    self.mock.get.return_value = policy
    issue_tracker = monorail.IssueTracker(IssueTrackerManager(project_name))

    self.testcase1.security_flag = True
    self.testcase1.security_severity = None
    self.testcase1.put()
    issue_filer.file_issue(self.testcase1, issue_tracker)
    self.assertIn('Security_Severity-High',
                  issue_tracker._itm.last_issue.labels)
    self.assertEqual(
        1,
        len(
            issue_tracker._itm.last_issue.get_labels_by_prefix(
                'Security_Severity-')))

  @parameterized.parameterized.expand([
      ('chromium', CHROMIUM_POLICY),
      ('oss-fuzz', OSS_FUZZ_POLICY),
  ])
  def test_security_severity_security_bug_severity_override(
      self, project_name, policy):
    """Test security severity label is set correct when testcase has its own
    severity but there is an override provided."""
    self.mock.get.return_value = policy
    issue_tracker = monorail.IssueTracker(IssueTrackerManager(project_name))

    self.testcase1.security_flag = True
    self.testcase1.security_severity = data_types.SecuritySeverity.HIGH
    self.testcase1.put()
    issue_filer.file_issue(
        self.testcase1,
        issue_tracker,
        security_severity=data_types.SecuritySeverity.MEDIUM)
    self.assertNotIn('Security_Severity-High',
                     issue_tracker._itm.last_issue.labels)
    self.assertIn('Security_Severity-Medium',
                  issue_tracker._itm.last_issue.labels)
    self.assertEqual(
        1,
        len(
            issue_tracker._itm.last_issue.get_labels_by_prefix(
                'Security_Severity-')))

  @parameterized.parameterized.expand([
      ('chromium', CHROMIUM_POLICY),
      ('oss-fuzz', OSS_FUZZ_POLICY),
  ])
  def test_security_severity_security_bug_with_severity_set(
      self, project_name, policy):
    """Test security severity label is set when testcase is a security bug and
    has a security severity."""
    self.mock.get.return_value = policy
    security_severity_string_map = {
        data_types.SecuritySeverity.HIGH: 'Security_Severity-High',
        data_types.SecuritySeverity.MEDIUM: 'Security_Severity-Medium',
        data_types.SecuritySeverity.LOW: 'Security_Severity-Low',
    }

    for security_severity in security_severity_string_map:
      issue_tracker = monorail.IssueTracker(IssueTrackerManager(project_name))

      self.testcase1.security_flag = True
      self.testcase1.security_severity = security_severity
      self.testcase1.put()

      issue_filer.file_issue(self.testcase1, issue_tracker)
      self.assertIn(security_severity_string_map[security_severity],
                    issue_tracker._itm.last_issue.labels)
      self.assertEqual(
          1,
          len(
              issue_tracker._itm.last_issue.get_labels_by_prefix(
                  'Security_Severity-')))

  @parameterized.parameterized.expand([
      ('chromium', CHROMIUM_POLICY),
      ('oss-fuzz', OSS_FUZZ_POLICY),
  ])
  def test_memory_tool_used(self, project_name, policy):
    """Test memory tool label is correctly set."""
    self.mock.get.return_value = policy
    for entry in issue_filer.MEMORY_TOOLS_LABELS:
      issue_tracker = monorail.IssueTracker(IssueTrackerManager(project_name))

      self.testcase1.crash_stacktrace = '\n\n%s\n' % entry['token']
      self.testcase1.put()
      issue_filer.file_issue(self.testcase1, issue_tracker)
      self.assertIn('Stability-' + entry['label'],
                    issue_tracker._itm.last_issue.labels)

  def test_reproducible_flag(self):
    """Test (un)reproducible flag is correctly set."""
    self.mock.get.return_value = CHROMIUM_POLICY
    issue_tracker = monorail.IssueTracker(IssueTrackerManager('chromium'))

    self.testcase1.one_time_crasher_flag = True
    self.testcase1.put()
    issue_filer.file_issue(self.testcase1, issue_tracker)
    self.assertIn('Unreproducible', issue_tracker._itm.last_issue.labels)
    self.assertCountEqual(['Unreproducible>Component'],
                          issue_tracker._itm.last_issue.components)

    self.testcase1.one_time_crasher_flag = False
    self.testcase1.put()
    issue_filer.file_issue(self.testcase1, issue_tracker)
    self.assertIn('Reproducible', issue_tracker._itm.last_issue.labels)

  def test_crash_labels(self):
    """Test crash label setting."""
    self.mock.get.return_value = CHROMIUM_POLICY
    issue_tracker = monorail.IssueTracker(IssueTrackerManager('chromium'))

    self.testcase1.crash_type = 'UNKNOWN'
    self.testcase1.put()
    issue_filer.file_issue(self.testcase1, issue_tracker)
    self.assertIn('Pri-1', issue_tracker._itm.last_issue.labels)
    self.assertIn('Stability-Crash', issue_tracker._itm.last_issue.labels)

    self.testcase1.crash_type = 'Undefined-shift'
    self.testcase1.put()
    issue_filer.file_issue(self.testcase1, issue_tracker)
    self.assertIn('Pri-2', issue_tracker._itm.last_issue.labels)
    self.assertNotIn('Stability-Crash', issue_tracker._itm.last_issue.labels)

  def test_footer_formatting(self):
    """Test footer message with formatting."""
    policy = issue_tracker_policy.IssueTrackerPolicy({
        'status': {
            'assigned': 'Assigned',
            'duplicate': 'Duplicate',
            'verified': 'Verified',
            'new': 'Untriaged',
            'wontfix': 'WontFix',
            'fixed': 'Fixed'
        },
        'all': {
            'status': 'new',
            'issue_body_footer': 'Target: %FUZZ_TARGET%, Project: %PROJECT%'
        },
        'non_security': {},
        'labels': {},
        'security': {},
        'existing': {},
    })
    self.mock.get.return_value = policy

    issue_tracker = monorail.IssueTracker(IssueTrackerManager('chromium'))
    self.testcase1.project_name = 'proj'
    self.testcase1.fuzzer_name = 'libFuzzer'
    self.testcase1.overridden_fuzzer_name = 'libFuzzer_proj_target'

    data_types.FuzzTarget(
        id='libFuzzer_proj_target',
        project='proj',
        engine='libFuzzer',
        binary='target').put()

    issue_filer.file_issue(self.testcase1, issue_tracker)
    self.assertIn('Target: target, Project: proj',
                  issue_tracker._itm.last_issue.body)


class MemoryToolLabelsTest(unittest.TestCase):
  """Memory tool labels tests."""
  DATA_DIRECTORY = os.path.join(os.path.dirname(__file__), 'issue_filer_data')

  def _read_test_data(self, name):
    """Helper function to read test data."""
    with open(os.path.join(self.DATA_DIRECTORY, name)) as handle:
      return handle.read()

  def test_memory_tools_labels_asan(self):
    """Run memory tools detection with test data."""
    expected_labels = ['Memory-AddressSanitizer']
    data = self._read_test_data('memory_tools_asan.txt')
    actual_labels = issue_filer.get_memory_tool_labels(data)

    self.assertEqual(actual_labels, expected_labels)

  def test_memory_tools_labels_asan_afl(self):
    """Run memory tools detection with test data."""
    expected_labels = ['Memory-AddressSanitizer', 'AFL']
    data = self._read_test_data('memory_tools_asan_afl.txt')
    actual_labels = issue_filer.get_memory_tool_labels(data)

    self.assertEqual(actual_labels, expected_labels)

  def test_memory_tools_labels_asan_libfuzzer(self):
    """Run memory tools detection with test data."""
    expected_labels = ['Memory-AddressSanitizer', 'LibFuzzer']
    data = self._read_test_data('memory_tools_asan_libfuzzer.txt')
    actual_labels = issue_filer.get_memory_tool_labels(data)

    self.assertEqual(actual_labels, expected_labels)

  def test_memory_tools_labels_asan_lsan(self):
    """Run memory tools detection with test data."""
    expected_labels = ['Memory-AddressSanitizer', 'Memory-LeakSanitizer']
    data = self._read_test_data('memory_tools_asan_lsan.txt')
    actual_labels = issue_filer.get_memory_tool_labels(data)

    self.assertEqual(actual_labels, expected_labels)

  def test_memory_tools_labels_msan(self):
    """Run memory tools detection with test data."""
    expected_labels = ['Memory-MemorySanitizer']
    data = self._read_test_data('memory_tools_msan.txt')
    actual_labels = issue_filer.get_memory_tool_labels(data)

    self.assertEqual(actual_labels, expected_labels)

  def test_memory_tools_labels_msan_libfuzzer(self):
    """Run memory tools detection with test data."""
    expected_labels = ['Memory-MemorySanitizer', 'LibFuzzer']
    data = self._read_test_data('memory_tools_msan_libfuzzer.txt')
    actual_labels = issue_filer.get_memory_tool_labels(data)

    self.assertEqual(actual_labels, expected_labels)

  def test_memory_tools_labels_tsan(self):
    """Run memory tools detection with test data."""
    expected_labels = ['ThreadSanitizer']
    data = self._read_test_data('memory_tools_tsan.txt')
    actual_labels = issue_filer.get_memory_tool_labels(data)

    self.assertEqual(actual_labels, expected_labels)

  def test_memory_tools_ubsan(self):
    """Run memory tools detection with test data."""
    expected_labels = ['UndefinedBehaviorSanitizer']
    data = self._read_test_data('memory_tools_ubsan.txt')
    actual_labels = issue_filer.get_memory_tool_labels(data)

    self.assertEqual(actual_labels, expected_labels)


@test_utils.with_cloud_emulators('datastore')
class UpdateImpactTest(unittest.TestCase):
  """Update impact tests."""

  def _make_mock_issue(self):
    mock_issue = mock.Mock(autospec=MonorailIssue)
    mock_issue.labels = LabelStore()

    return mock_issue

  def setUp(self):
    helpers.patch_environ(self)
    self.testcase = data_types.Testcase()
    self.testcase.one_time_crasher_flag = False
    self.testcase.crash_state = 'fake_crash'

  def test_update_impact_stable_from_regression(self):
    """Tests updating impact to Stable from the regression range."""
    self.testcase.regression = '0:1000'
    mock_issue = self._make_mock_issue()

    issue_filer.update_issue_impact_labels(self.testcase, mock_issue)
    six.assertCountEqual(self, ['Security_Impact-Extended'],
                         mock_issue.labels.added)
    six.assertCountEqual(self, [], mock_issue.labels.removed)

  def test_update_impact_extended_stable(self):
    """Tests updating impact to Extended Stable."""
    self.testcase.is_impact_set_flag = True
    self.testcase.impact_extended_stable_version = '99.1024.11.42'

    mock_issue = self._make_mock_issue()

    issue_filer.update_issue_impact_labels(self.testcase, mock_issue)
    six.assertCountEqual(self, ['Security_Impact-Extended', 'FoundIn-99'],
                         mock_issue.labels.added)
    six.assertCountEqual(self, [], mock_issue.labels.removed)

  def test_update_impact_stable(self):
    """Tests updating impact to Stable."""
    self.testcase.is_impact_set_flag = True
    self.testcase.impact_stable_version = '99.1024.11.42'

    mock_issue = self._make_mock_issue()

    issue_filer.update_issue_impact_labels(self.testcase, mock_issue)
    six.assertCountEqual(self, ['Security_Impact-Stable', 'FoundIn-99'],
                         mock_issue.labels.added)
    six.assertCountEqual(self, [], mock_issue.labels.removed)

  def test_update_impact_beta(self):
    """Tests updating impact to Beta."""
    self.testcase.is_impact_set_flag = True
    self.testcase.impact_beta_version = '100.1044.44.44'

    mock_issue = self._make_mock_issue()

    issue_filer.update_issue_impact_labels(self.testcase, mock_issue)
    six.assertCountEqual(self, ['Security_Impact-Beta', 'FoundIn-100'],
                         mock_issue.labels.added)
    six.assertCountEqual(self, [], mock_issue.labels.removed)

  def test_update_impact_head(self):
    """Tests updating impact to Head."""
    self.testcase.is_impact_set_flag = True

    mock_issue = self._make_mock_issue()

    issue_filer.update_issue_impact_labels(self.testcase, mock_issue)
    six.assertCountEqual(self, ['Security_Impact-Head'],
                         mock_issue.labels.added)
    six.assertCountEqual(self, [], mock_issue.labels.removed)

  def test_no_impact_for_unreproducible_testcase(self):
    """Tests no impact for unreproducible testcase on trunk and which also
    does not crash on stable and beta."""
    self.testcase.is_impact_set_flag = True
    self.testcase.crash_state = ''

    mock_issue = self._make_mock_issue()

    issue_filer.update_issue_impact_labels(self.testcase, mock_issue)
    six.assertCountEqual(self, [], mock_issue.labels.added)
    six.assertCountEqual(self, [], mock_issue.labels.removed)

  def test_no_impact_if_not_set(self):
    """Tests no impact if the impact flag is not set."""
    mock_issue = self._make_mock_issue()

    issue_filer.update_issue_impact_labels(self.testcase, mock_issue)
    six.assertCountEqual(self, [], mock_issue.labels.added)
    six.assertCountEqual(self, [], mock_issue.labels.removed)

  def test_replace_impact(self):
    """Tests replacing impact."""
    self.testcase.is_impact_set_flag = True

    mock_issue = self._make_mock_issue()
    mock_issue.labels.add('Security_Impact-Beta')
    mock_issue.labels.reset_tracking()

    issue_filer.update_issue_impact_labels(self.testcase, mock_issue)
    six.assertCountEqual(self, ['Security_Impact-Head'],
                         mock_issue.labels.added)
    six.assertCountEqual(self, ['Security_Impact-Beta'],
                         mock_issue.labels.removed)

  def test_replace_same_impact(self):
    """Tests replacing same impact."""
    self.testcase.is_impact_set_flag = True

    mock_issue = self._make_mock_issue()
    mock_issue.labels.add('Security_Impact-Head')
    mock_issue.labels.reset_tracking()

    issue_filer.update_issue_impact_labels(self.testcase, mock_issue)
    six.assertCountEqual(self, [], mock_issue.labels.added)
    six.assertCountEqual(self, [], mock_issue.labels.removed)

  def test_component_add_label(self):
    """Test that we set labels for component builds."""
    self.testcase.job_type = 'job'
    self.testcase.impact_extended_stable_version = '1.2.3.4'
    self.testcase.impact_stable_version = '2.3.4.5'
    self.testcase.impact_beta_version = '3.4.5.6'
    self.testcase.put()

    data_types.Job(
        name='job',
        environment_string=(
            'RELEASE_BUILD_BUCKET_PATH = '
            'https://example.com/blah-v8-component-([0-9]+).zip\n')).put()

    self.testcase.is_impact_set_flag = True
    mock_issue = self._make_mock_issue()
    issue_filer.update_issue_impact_labels(self.testcase, mock_issue)
    six.assertCountEqual(
        self,
        ['Security_Impact-Extended', 'FoundIn-1', 'FoundIn-2', 'FoundIn-3'],
        mock_issue.labels.added)
    six.assertCountEqual(self, [], mock_issue.labels.removed)


@test_utils.with_cloud_emulators('datastore', 'pubsub')
class NotifyIssueUpdateTests(unittest.TestCase):
  """notify_issue_update tests."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.config.local_config.ProjectConfig',
    ])

    self.topic = 'projects/project/topics/issue-updates'
    self.subscription = 'projects/project/subscriptions/issue-updates'

    self.mock.ProjectConfig.return_value = mock_config.MockConfig({
        'issue_updates': {
            'pubsub_topic': self.topic,
        },
    })

    self.pubsub_client = pubsub.PubSubClient()
    self.pubsub_client.create_topic(self.topic)
    self.pubsub_client.create_subscription(self.subscription, self.topic)

    self.testcase = data_types.Testcase(
        crash_address='0xffff',
        security_flag=True,
        crash_type='CRASH TYPE',
        crash_state='CRASH STATE',
        bug_information='123')
    self.testcase.put()

  def test_basic(self):
    """Basic test."""
    issue_filer.notify_issue_update(self.testcase, 'new')
    messages = self.pubsub_client.pull_from_subscription(
        self.subscription, max_messages=16)
    self.assertEqual(1, len(messages))
    self.assertDictEqual({
        'crash_address': '0xffff',
        'crash_state': 'CRASH STATE',
        'crash_type': 'CRASH TYPE',
        'issue_id': '123',
        'security': 'true',
        'status': 'new',
        'testcase_id': '1'
    }, messages[0].attributes)

  def test_no_topic(self):
    """Test when no topic is specified."""
    self.mock.ProjectConfig.return_value = mock_config.MockConfig({})
    issue_filer.notify_issue_update(self.testcase, 'new')
    messages = self.pubsub_client.pull_from_subscription(
        self.subscription, max_messages=16)
    self.assertEqual(0, len(messages))

  def test_no_issue(self):
    """Test no issue id."""
    self.testcase.bug_information = None
    issue_filer.notify_issue_update(self.testcase, 'new')
    messages = self.pubsub_client.pull_from_subscription(
        self.subscription, max_messages=16)
    self.assertEqual(1, len(messages))
    self.assertDictEqual({
        'crash_address': '0xffff',
        'crash_state': 'CRASH STATE',
        'crash_type': 'CRASH TYPE',
        'issue_id': '',
        'security': 'true',
        'status': 'new',
        'testcase_id': '1'
    }, messages[0].attributes)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler used for loading bigquery data."""

import datetime
import random
import time

from googleapiclient.errors import HttpError
import httplib2

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import big_query
from clusterfuzz._internal.metrics import fuzzer_stats
from clusterfuzz._internal.metrics import fuzzer_stats_schema
from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import handler

STATS_KINDS = [fuzzer_stats.JobRun, fuzzer_stats.TestcaseRun]

NUM_RETRIES = 2
RETRY_SLEEP_TIME = 5


class Handler(base_handler.Handler):
  """Cron handler for loading bigquery stats."""

  def _utc_now(self):
    """Return datetime.datetime.utcnow()."""
    return datetime.datetime.utcnow()

  def _execute_insert_request(self, request):
    """Executes a table/dataset insert request, retrying on transport errors."""
    for i in range(NUM_RETRIES + 1):
      try:
        request.execute()
        return True
      except HttpError as e:
        if e.resp.status == 409:
          # Already exists.
          return True

        logs.log_error('Failed to insert table/dataset.')
        return False
      except httplib2.HttpLib2Error:
        # Transport error.
        time.sleep(random.uniform(0, (1 << i) * RETRY_SLEEP_TIME))
        continue

    logs.log_error('Failed to insert table/dataset.')
    return False

  def _create_dataset_if_needed(self, bigquery, dataset_id):
    """Create a new dataset if necessary."""
    project_id = utils.get_application_id()
    dataset_body = {
        'datasetReference': {
            'datasetId': dataset_id,
            'projectId': project_id,
        },
    }
    dataset_insert = bigquery.datasets().insert(
        projectId=project_id, body=dataset_body)

    return self._execute_insert_request(dataset_insert)

  def _create_table_if_needed(self, bigquery, dataset_id, table_id):
    """Create a new table if needed."""
    project_id = utils.get_application_id()
    table_body = {
        'tableReference': {
            'datasetId': dataset_id,
            'projectId': project_id,
            'tableId': table_id,
        },
        'timePartitioning': {
            'type': 'DAY',
        },
    }

    table_insert = bigquery.tables().insert(
        projectId=project_id, datasetId=dataset_id, body=table_body)
    return self._execute_insert_request(table_insert)

  def _load_data(self, bigquery, fuzzer):
    """Load yesterday's stats into BigQuery."""
    project_id = utils.get_application_id()

    yesterday = (self._utc_now().date() - datetime.timedelta(days=1))
    date_string = yesterday.strftime('%Y%m%d')
    timestamp = utils.utc_date_to_timestamp(yesterday)

    dataset_id = fuzzer_stats.dataset_name(fuzzer)
    if not self._create_dataset_if_needed(bigquery, dataset_id):
      return

    for kind in STATS_KINDS:
      kind_name = kind.__name__
      table_id = kind_name
      if not self._create_table_if_needed(bigquery, dataset_id, table_id):
        continue

      if kind == fuzzer_stats.TestcaseRun:
        schema = fuzzer_stats_schema.get(fuzzer)
      else:
        schema = kind.SCHEMA

      gcs_path = fuzzer_stats.get_gcs_stats_path(kind_name, fuzzer, timestamp)
      load = {
          'destinationTable': {
              'projectId': project_id,
              'tableId': table_id + '$' + date_string,
              'datasetId': dataset_id,
          },
          'schemaUpdateOptions': ['ALLOW_FIELD_ADDITION',],
          'sourceFormat': 'NEWLINE_DELIMITED_JSON',
          'sourceUris': ['gs:/' + gcs_path + '*.json'],
          'writeDisposition': 'WRITE_TRUNCATE',
      }
      if schema is not None:
        load['schema'] = schema

      job_body = {
          'configuration': {
              'load': load,
          },
      }

      logs.log("Uploading job to BigQuery.", job_body=job_body)
      request = bigquery.jobs().insert(projectId=project_id, body=job_body)
      response = request.execute()

      # We cannot really check the response here, as the query might be still
      # running, but having a BigQuery jobId in the log would make our life
      # simpler if we ever have to manually check the status of the query.
      # See https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs/query.
      logs.log('Response from BigQuery: %s' % response)

  @handler.cron()
  def get(self):
    """Load bigquery stats from GCS."""
    if not big_query.get_bucket():
      logs.log_error('Loading stats to BigQuery failed: missing bucket name.')
      return

    # Retrieve list of fuzzers before iterating them, since the query can expire
    # as we create the load jobs.
    bigquery_client = big_query.get_api_client()
    for fuzzer in list(data_types.Fuzzer.query()):
      logs.log('Loading stats to BigQuery for %s.' % fuzzer.name)
      self._load_data(bigquery_client, fuzzer.name)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Task queue functions."""

import contextlib
import datetime
import random
import threading
import time

from clusterfuzz._internal.base import external_tasks
from clusterfuzz._internal.base import persistent_cache
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.fuzzing import fuzzer_selection
from clusterfuzz._internal.google_cloud_utils import pubsub
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

# Task queue prefixes for various job types.
JOBS_PREFIX = 'jobs'
HIGH_END_JOBS_PREFIX = 'high-end-jobs'

# Default task queue names for various job types. These will be different for
# different platforms with the platform name added as suffix later.
JOBS_TASKQUEUE = JOBS_PREFIX
HIGH_END_JOBS_TASKQUEUE = HIGH_END_JOBS_PREFIX

# ML job is currently supported on Linux only.
ML_JOBS_TASKQUEUE = 'ml-jobs-linux'

# Limits on number of tasks leased at once and in total.
MAX_LEASED_TASKS_LIMIT = 1000
MAX_TASKS_LIMIT = 100000

# Various variables for task leasing and completion times (in seconds).
TASK_COMPLETION_BUFFER = 90 * 60
TASK_CREATION_WAIT_INTERVAL = 2 * 60
TASK_EXCEPTION_WAIT_INTERVAL = 5 * 60
TASK_LEASE_SECONDS = 6 * 60 * 60  # Can be overridden via environment variable.
TASK_LEASE_SECONDS_BY_COMMAND = {
    'corpus_pruning': 24 * 60 * 60,
    'regression': 24 * 60 * 60,
}

TASK_QUEUE_DISPLAY_NAMES = {
    'LINUX': 'Linux',
    'LINUX_WITH_GPU': 'Linux (with GPU)',
    'LINUX_UNTRUSTED': 'Linux (untrusted)',
    'ANDROID': 'Android',
    'ANDROID_KERNEL': 'Android Kernel',
    'ANDROID_KERNEL_X86': 'Android Kernel (X86)',
    'ANDROID_AUTO': 'Android Auto',
    'ANDROID_X86': 'Android (x86)',
    'ANDROID_EMULATOR': 'Android (Emulated)',
    'CHROMEOS': 'Chrome OS',
    'FUCHSIA': 'Fuchsia OS',
    'MAC': 'Mac',
    'WINDOWS': 'Windows',
    'WINDOWS_WITH_GPU': 'Windows (with GPU)',
}

VALID_REDO_TASKS = ['minimize', 'regression', 'progression', 'impact', 'blame']

LEASE_FAIL_WAIT = 10
LEASE_RETRIES = 5

TASK_PAYLOAD_KEY = 'task_payload'
TASK_END_TIME_KEY = 'task_end_time'


class Error(Exception):
  """Base exception class."""


class InvalidRedoTask(Error):

  def __init__(self, task):
    super().__init__("The task '%s' is invalid." % task)


def queue_suffix_for_platform(platform):
  """Get the queue suffix for a platform."""
  return '-' + platform.lower().replace('_', '-')


def default_queue_suffix():
  """Get the queue suffix for the current platform."""
  queue_override = environment.get_value('QUEUE_OVERRIDE')
  if queue_override:
    return queue_suffix_for_platform(queue_override)

  return queue_suffix_for_platform(environment.platform())


def regular_queue(prefix=JOBS_PREFIX):
  """Get the regular jobs queue."""
  return prefix + default_queue_suffix()


def high_end_queue():
  """Get the high end jobs queue."""
  return regular_queue(prefix=HIGH_END_JOBS_PREFIX)


def default_queue():
  """Get the default jobs queue."""
  thread_multiplier = environment.get_value('THREAD_MULTIPLIER')
  if thread_multiplier and thread_multiplier > 1:
    return high_end_queue()

  return regular_queue()


def get_command_override():
  """Get command override task."""
  command_override = environment.get_value('COMMAND_OVERRIDE', '').strip()
  if not command_override:
    return None

  parts = command_override.split()
  if len(parts) != 3:
    raise ValueError('Command override should have 3 components.')

  return Task(*parts, is_command_override=True)


def get_fuzz_task():
  """Try to get a fuzz task."""
  argument, job = fuzzer_selection.get_fuzz_task_payload()
  if not argument:
    return None

  return Task('fuzz', argument, job)


def get_high_end_task():
  """Get a high end task."""
  task = get_regular_task(queue=high_end_queue())
  if not task:
    return None

  task.high_end = True
  return task


def get_regular_task(queue=None):
  """Get a regular task."""
  if not queue:
    queue = regular_queue()

  pubsub_client = pubsub.PubSubClient()
  application_id = utils.get_application_id()
  while True:
    messages = pubsub_client.pull_from_subscription(
        pubsub.subscription_name(application_id, queue), max_messages=1)

    if not messages:
      return None

    try:
      task = PubSubTask(messages[0])
    except KeyError:
      logs.log_error('Received an invalid task, discarding...')
      messages[0].ack()
      continue

    # Check that this task should be run now (past the ETA). Otherwise we defer
    # its execution.
    if not task.defer():
      return task


def get_task():
  """Get a task."""
  task = get_command_override()
  if task:
    return task

  # TODO(unassigned): Remove this hack.
  if environment.get_value('ML'):
    return get_regular_task(queue=ML_JOBS_TASKQUEUE)

  allow_all_tasks = not environment.get_value('PREEMPTIBLE')
  if allow_all_tasks:
    # Check the high-end jobs queue for bots with multiplier greater than 1.
    thread_multiplier = environment.get_value('THREAD_MULTIPLIER')
    if thread_multiplier and thread_multiplier > 1:
      task = get_high_end_task()
      if task:
        return task

    task = get_regular_task()
    if task:
      return task

  task = get_fuzz_task()
  if not task:
    logs.log_error('Failed to get any fuzzing tasks. This should not happen.')
    time.sleep(TASK_EXCEPTION_WAIT_INTERVAL)

  return task


class Task(object):
  """Represents a task."""

  def __init__(self,
               command,
               argument,
               job,
               eta=None,
               is_command_override=False,
               high_end=False):
    self.command = command
    self.argument = argument
    self.job = job
    self.eta = eta
    self.is_command_override = is_command_override
    self.high_end = high_end

  def attribute(self, _):
    return None

  def payload(self):
    """Get the payload."""
    return ' '.join([self.command, self.argument, self.job])

  def to_pubsub_message(self):
    """Convert the task to a pubsub message."""
    attributes = {
        'command': self.command,
        'argument': str(self.argument),
        'job': self.job,
    }

    if self.eta:
      attributes['eta'] = str(utils.utc_datetime_to_timestamp(self.eta))

    return pubsub.Message(attributes=attributes)

  @contextlib.contextmanager
  def lease(self):
    """Maintain a lease for the task. Track only start and end by default."""
    # Assume default time for non-pubsub tasks.
    track_task_start(self, TASK_LEASE_SECONDS)
    yield
    track_task_end()


class PubSubTask(Task):
  """A Pub/Sub task."""

  def __init__(self, pubsub_message):
    self._pubsub_message = pubsub_message
    super().__init__(
        self.attribute('command'), self.attribute('argument'),
        self.attribute('job'))

    self.eta = datetime.datetime.utcfromtimestamp(float(self.attribute('eta')))

  def attribute(self, key):
    """Return attribute value."""
    return self._pubsub_message.attributes[key]

  def defer(self):
    """Defer a task until its ETA. Returns whether or not we deferred."""
    now = utils.utcnow()
    if now >= self.eta:
      return False

    # Extend the deadline until the ETA, or MAX_ACK_DEADLINE.
    time_until_eta = int((self.eta - now).total_seconds())
    logs.log('Deferring task "%s".' % self.payload())
    self._pubsub_message.modify_ack_deadline(
        min(pubsub.MAX_ACK_DEADLINE, time_until_eta))
    return True

  @contextlib.contextmanager
  def lease(self, _event=None):  # pylint: disable=arguments-differ
    """Maintain a lease for the task."""
    task_lease_timeout = TASK_LEASE_SECONDS_BY_COMMAND.get(
        self.command, get_task_lease_timeout())

    environment.set_value('TASK_LEASE_SECONDS', task_lease_timeout)
    track_task_start(self, task_lease_timeout)

    if _event is None:
      _event = threading.Event()

    leaser_thread = _PubSubLeaserThread(self._pubsub_message, _event,
                                        task_lease_timeout)
    leaser_thread.start()
    try:
      yield leaser_thread
    finally:
      _event.set()
      leaser_thread.join()

    # If we get here the task succeeded in running. Acknowledge the message.
    self._pubsub_message.ack()
    track_task_end()


class _PubSubLeaserThread(threading.Thread):
  """Thread that continuously renews the lease for a message."""

  EXTENSION_TIME_SECONDS = 10 * 60  # 10 minutes.

  def __init__(self, message, done_event, max_lease_seconds):
    super().__init__()

    self.daemon = True
    self._message = message
    self._done_event = done_event
    self._max_lease_seconds = max_lease_seconds

  def run(self):
    """Run the leaser thread."""
    latest_end_time = time.time() + self._max_lease_seconds

    while True:
      try:
        time_left = latest_end_time - time.time()
        if time_left <= 0:
          logs.log('Lease reached maximum lease time of {} seconds, '
                   'stopping renewal.'.format(self._max_lease_seconds))
          break

        extension_seconds = min(self.EXTENSION_TIME_SECONDS, time_left)

        logs.log(
            'Renewing lease for task by {} seconds.'.format(extension_seconds))
        self._message.modify_ack_deadline(extension_seconds)

        # Schedule renewals earlier than the extension to avoid race conditions
        # and performing the next extension too late.
        wait_seconds = min(time_left, self.EXTENSION_TIME_SECONDS // 2)

        # Wait until the next scheduled renewal, or if the task is complete.
        if self._done_event.wait(wait_seconds):
          logs.log('Task complete, stopping renewal.')
          break
      except Exception:
        logs.log_error('Leaser thread failed.')


def add_task(command, argument, job_type, queue=None, wait_time=None):
  """Add a new task to the job queue."""
  # Old testcases may pass in queue=None explicitly,
  # so we must check this here.
  if not queue:
    queue = default_queue()

  if wait_time is None:
    wait_time = random.randint(1, TASK_CREATION_WAIT_INTERVAL)

  if job_type != 'none':
    job = data_types.Job.query(data_types.Job.name == job_type).get()
    if not job:
      raise Error(f'Job {job_type} not found.')

    if job.is_external():
      external_tasks.add_external_task(command, argument, job)
      return

  # Add the task.
  eta = utils.utcnow() + datetime.timedelta(seconds=wait_time)
  task = Task(command, argument, job_type, eta=eta)
  pubsub_client = pubsub.PubSubClient()
  pubsub_client.publish(
      pubsub.topic_name(utils.get_application_id(), queue),
      [task.to_pubsub_message()])


def get_task_lease_timeout():
  """Return the task lease timeout."""
  return environment.get_value('TASK_LEASE_SECONDS', TASK_LEASE_SECONDS)


def get_task_completion_deadline():
  """Return task completion deadline. This gives an additional buffer over the
  task lease deadline."""
  start_time = time.time()
  task_lease_timeout = get_task_lease_timeout()
  return start_time + task_lease_timeout - TASK_COMPLETION_BUFFER


def queue_for_platform(platform, is_high_end=False):
  """Return the queue for the platform."""
  prefix = HIGH_END_JOBS_PREFIX if is_high_end else JOBS_PREFIX
  return prefix + queue_suffix_for_platform(platform)


def queue_for_testcase(testcase):
  """Return the right queue for the testcase."""
  is_high_end = (
      testcase.queue and testcase.queue.startswith(HIGH_END_JOBS_PREFIX))
  return queue_for_job(testcase.job_type, is_high_end=is_high_end)


def queue_for_job(job_name, is_high_end=False):
  """Queue for job."""
  job = data_types.Job.query(data_types.Job.name == job_name).get()
  if not job:
    raise Error('Job {} not found.'.format(job_name))

  return queue_for_platform(job.platform, is_high_end)


def redo_testcase(testcase, tasks, user_email):
  """Redo specific tasks for a testcase."""
  for task in tasks:
    if task not in VALID_REDO_TASKS:
      raise InvalidRedoTask(task)

  minimize = 'minimize' in tasks
  regression = 'regression' in tasks
  progression = 'progression' in tasks
  impact = 'impact' in tasks
  blame = 'blame' in tasks

  task_list = []
  testcase_id = testcase.key.id()

  # Metadata keys to clear based on which redo tasks were selected.
  metadata_keys_to_clear = ['potentially_flaky']

  if minimize:
    task_list.append('minimize')
    testcase.minimized_keys = ''
    testcase.set_metadata('redo_minimize', True, update_testcase=False)
    metadata_keys_to_clear += [
        'env', 'current_minimization_phase_attempts', 'minimization_phase'
    ]

    # If this testcase was archived during minimization, update the state.
    testcase.archive_state &= ~data_types.ArchiveStatus.MINIMIZED

  if regression:
    task_list.append('regression')
    testcase.regression = ''
    metadata_keys_to_clear += ['last_regression_min', 'last_regression_max']

  if progression:
    task_list.append('progression')
    testcase.fixed = ''
    testcase.open = True
    testcase.last_tested_crash_stacktrace = None
    testcase.triaged = False
    testcase.set_metadata('progression_pending', True, update_testcase=False)
    metadata_keys_to_clear += [
        'last_progression_min', 'last_progression_max', 'last_tested_revision'
    ]

  if impact:
    task_list.append('impact')
    testcase.is_impact_set_flag = False

  if blame:
    task_list.append('blame')
    testcase.set_metadata('blame_pending', True, update_testcase=False)
    testcase.set_metadata('predator_result', None, update_testcase=False)

  for key in metadata_keys_to_clear:
    testcase.delete_metadata(key, update_testcase=False)

  testcase.comments += '[%s] %s: Redo task(s): %s\n' % (
      utils.current_date_time(), user_email, ', '.join(sorted(task_list)))
  testcase.one_time_crasher_flag = False
  testcase.put()

  # Allow new notifications to be sent for this testcase.
  notifications = ndb_utils.get_all_from_query(
      data_types.Notification.query(
          data_types.Notification.testcase_id == testcase.key.id()),
      keys_only=True)
  ndb_utils.delete_multi(notifications)

  # If we are re-doing minimization, other tasks will be done automatically
  # after minimization completes. So, don't add those tasks.
  if minimize:
    add_task('minimize', testcase_id, testcase.job_type,
             queue_for_testcase(testcase))
  else:
    if regression:
      add_task('regression', testcase_id, testcase.job_type,
               queue_for_testcase(testcase))

    if progression:
      add_task('progression', testcase_id, testcase.job_type,
               queue_for_testcase(testcase))

    if impact:
      add_task('impact', testcase_id, testcase.job_type,
               queue_for_testcase(testcase))

    if blame:
      add_task('blame', testcase_id, testcase.job_type,
               queue_for_testcase(testcase))


def get_task_payload():
  """Return current task payload."""
  return persistent_cache.get_value(TASK_PAYLOAD_KEY)


def get_task_end_time():
  """Return current task end time."""
  return persistent_cache.get_value(
      TASK_END_TIME_KEY, constructor=datetime.datetime.utcfromtimestamp)


def track_task_start(task, task_duration):
  """Cache task information."""
  persistent_cache.set_value(TASK_PAYLOAD_KEY, task.payload())
  persistent_cache.set_value(TASK_END_TIME_KEY, time.time() + task_duration)

  # Don't wait on |run_heartbeat|, update task information as soon as it starts.
  from clusterfuzz._internal.datastore import data_handler
  data_handler.update_heartbeat(force_update=True)


def track_task_end():
  """Remove cached task information."""
  persistent_cache.delete_value(TASK_PAYLOAD_KEY)
  persistent_cache.delete_value(TASK_END_TIME_KEY)

  # Don't wait on |run_heartbeat|, remove task information as soon as it ends.
  from clusterfuzz._internal.datastore import data_handler
  data_handler.update_heartbeat(force_update=True)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions for app-specific trials/experiments."""

import random

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment


class Trials:
  """Helper class for selecting app-specific extra flags."""

  def __init__(self):
    self.trials = []

    app_name = environment.get_value('APP_NAME')
    if not app_name:
      return

    # Convert the app_name to lowercase. Case may vary by platform.
    app_name = app_name.lower()

    # Hack: strip file extensions that may be appended on various platforms.
    extensions_to_strip = ['.exe', '.apk']
    for extension in extensions_to_strip:
      app_name = utils.strip_from_right(app_name, extension)

    self.trials = list(
        data_types.Trial.query(data_types.Trial.app_name == app_name))

  def setup_additional_args_for_app(self):
    """Select additional args for the specified app at random."""
    trial_args = [
        trial.app_args
        for trial in self.trials
        if random.random() < trial.probability
    ]
    if not trial_args:
      return

    trial_app_args = ' '.join(trial_args)
    app_args = environment.get_value('APP_ARGS', '')
    environment.set_value('APP_ARGS', '%s %s' % (app_args, trial_app_args))
    environment.set_value('TRIAL_APP_ARGS', trial_app_args)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions to file issues."""

import itertools
import re

from clusterfuzz._internal.base import external_users
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.crash_analysis import severity_analyzer
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import pubsub
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from libs.issue_management import issue_tracker_policy

NON_CRASH_TYPES = [
    'Data race',
    'Direct-leak',
    'Float-cast-overflow',
    'Incorrect-function-pointer-type',
    'Integer-overflow',
    'Non-positive-vla-bound-value',
    'RUNTIME_ASSERT',
    'Undefined-shift',
    'Unsigned-integer-overflow',
]

MEMORY_TOOLS_LABELS = [
    {
        'token': 'AddressSanitizer',
        'label': 'Memory-AddressSanitizer'
    },
    {
        'token': 'LeakSanitizer',
        'label': 'Memory-LeakSanitizer'
    },
    {
        'token': 'MemorySanitizer',
        'label': 'Memory-MemorySanitizer'
    },
    {
        'token': 'ThreadSanitizer',
        'label': 'ThreadSanitizer'
    },
    {
        'token': 'UndefinedBehaviorSanitizer',
        'label': 'UndefinedBehaviorSanitizer'
    },
    {
        'token': 'afl',
        'label': 'AFL'
    },
    {
        'token': 'libfuzzer',
        'label': 'LibFuzzer'
    },
]

STACKFRAME_LINE_REGEX = re.compile(r'\s*#\d+\s+0x[0-9A-Fa-f]+\s*')


def platform_substitution(label, testcase, _):
  """Platform substitution."""
  platform = None
  if environment.is_chromeos_job(testcase.job_type):
    # ChromeOS fuzzers run on Linux platform, so use correct OS-Chrome for
    # tracking.
    platform = 'Chrome'
  elif testcase.platform_id:
    platform = testcase.platform_id.split(':')[0].capitalize()

  if not platform:
    return []

  return [label.replace('%PLATFORM%', platform)]


def current_date():
  """Date format."""
  return utils.utcnow().date().isoformat()


def date_substitution(label, *_):
  """Date substitution."""
  return [label.replace('%YYYY-MM-DD%', current_date())]


def sanitizer_substitution(label, testcase, _):
  """Sanitizer substitution."""
  stacktrace = data_handler.get_stacktrace(testcase)
  memory_tool_labels = get_memory_tool_labels(stacktrace)

  return [
      label.replace('%SANITIZER%', memory_tool)
      for memory_tool in memory_tool_labels
  ]


def severity_substitution(label, testcase, security_severity):
  """Severity substitution."""
  # Use severity from testcase if one is not available.
  if security_severity is None:
    security_severity = testcase.security_severity

  # Set to default high severity if we can't determine it automatically.
  if not data_types.SecuritySeverity.is_valid(security_severity):
    security_severity = data_types.SecuritySeverity.HIGH

  security_severity_string = severity_analyzer.severity_to_string(
      security_severity)
  return [label.replace('%SEVERITY%', security_severity_string)]


def impact_to_string(impact):
  """Convert an impact value to a human-readable string."""
  impact_map = {
      data_types.SecurityImpact.EXTENDED_STABLE: 'Extended',
      data_types.SecurityImpact.STABLE: 'Stable',
      data_types.SecurityImpact.BETA: 'Beta',
      data_types.SecurityImpact.HEAD: 'Head',
      data_types.SecurityImpact.NONE: 'None',
      data_types.SecurityImpact.MISSING: data_types.MISSING_VALUE_STRING,
  }

  return impact_map[impact]


def _get_impact_from_labels(labels):
  """Get the impact from the label list."""
  labels = [label.lower() for label in labels]
  if 'security_impact-extended' in labels:
    return data_types.SecurityImpact.EXTENDED_STABLE
  if 'security_impact-stable' in labels:
    return data_types.SecurityImpact.STABLE
  if 'security_impact-beta' in labels:
    return data_types.SecurityImpact.BETA
  if 'security_impact-head' in labels:
    return data_types.SecurityImpact.HEAD
  if 'security_impact-none' in labels:
    return data_types.SecurityImpact.NONE
  return data_types.SecurityImpact.MISSING


def update_issue_impact_labels(testcase, issue):
  """Update impact labels on issue."""
  if testcase.one_time_crasher_flag:
    return

  existing_impact = _get_impact_from_labels(issue.labels)

  if testcase.regression.startswith('0:'):
    # If the regression range starts from the start of time,
    # then we assume that the bug impacts stable.
    new_impact = data_types.SecurityImpact.EXTENDED_STABLE
  elif testcase.is_impact_set_flag:
    # Add impact label based on testcase's impact value.
    if testcase.impact_extended_stable_version:
      new_impact = data_types.SecurityImpact.EXTENDED_STABLE
    elif testcase.impact_stable_version:
      new_impact = data_types.SecurityImpact.STABLE
    elif testcase.impact_beta_version:
      new_impact = data_types.SecurityImpact.BETA
    elif testcase.is_crash():
      new_impact = data_types.SecurityImpact.HEAD
    else:
      # Testcase is unreproducible and does not impact extended stable, stable
      # and beta branches. In this case, there is no impact information.
      return
  else:
    # No impact information.
    return

  update_issue_foundin_labels(testcase, issue)

  if existing_impact == new_impact:
    # Correct impact already set.
    return

  if existing_impact != data_types.SecurityImpact.MISSING:
    issue.labels.remove('Security_Impact-' + impact_to_string(existing_impact))

  issue.labels.add('Security_Impact-' + impact_to_string(new_impact))


def update_issue_foundin_labels(testcase, issue):
  """Updates FoundIn- labels on issue."""
  if not testcase.is_impact_set_flag:
    return
  versions_foundin = [
      x for x in [
          testcase.impact_beta_version, testcase.impact_stable_version,
          testcase.impact_extended_stable_version, testcase.impact_head_version
      ] if x
  ]
  milestones_foundin = {x.split('.')[0] for x in versions_foundin}
  for found_milestone in milestones_foundin:
    if f'foundin-{found_milestone}' in issue.labels:
      continue
    issue.labels.add('FoundIn-' + found_milestone)


def apply_substitutions(policy, label, testcase, security_severity=None):
  """Apply label substitutions."""
  if label is None:
    # If the label is not configured, then nothing to subsitute.
    return []

  label_substitutions = (
      ('%PLATFORM%', platform_substitution),
      ('%YYYY-MM-DD%', date_substitution),
      ('%SANITIZER%', sanitizer_substitution),
      ('%SEVERITY%', severity_substitution),
  )

  for marker, handler in label_substitutions:
    if marker in label:
      return [
          policy.substitution_mapping(label)
          for label in handler(label, testcase, security_severity)
      ]

  # No match found. Return unmodified label.
  return [label]


def get_label_pattern(label):
  """Get the label pattern regex."""
  return re.compile('^' + re.sub(r'%.*?%', r'(.*)', label) + '$', re.IGNORECASE)


def get_memory_tool_labels(stacktrace):
  """Distinguish memory tools used and return corresponding labels."""
  # Remove stack frames and paths to source code files. This helps to avoid
  # confusion when function names or source paths contain a memory tool token.
  data = ''
  for line in stacktrace.split('\n'):
    if STACKFRAME_LINE_REGEX.match(line):
      continue
    data += line + '\n'

  labels = [t['label'] for t in MEMORY_TOOLS_LABELS if t['token'] in data]
  return labels


def _get_from_metadata(testcase, name):
  """Get values from testcase metadata."""
  return utils.parse_delimited(
      testcase.get_metadata(name, ''),
      delimiter=',',
      strip=True,
      remove_empty=True)


def notify_issue_update(testcase, status):
  """Notify that an issue update occurred (i.e. issue was filed or closed)."""
  topic = local_config.ProjectConfig().get('issue_updates.pubsub_topic')
  if not topic:
    return

  pubsub_client = pubsub.PubSubClient()
  pubsub_client.publish(
      topic, [
          pubsub.Message(
              attributes={
                  'crash_address': testcase.crash_address,
                  'crash_state': testcase.crash_state,
                  'crash_type': testcase.crash_type,
                  'issue_id': testcase.bug_information or '',
                  'security': str(testcase.security_flag).lower(),
                  'status': status,
                  'testcase_id': str(testcase.key.id()),
              })
      ])


def file_issue(testcase,
               issue_tracker,
               security_severity=None,
               user_email=None,
               additional_ccs=None):
  """File an issue for the given test case."""
  logs.log('Filing new issue for testcase: %d' % testcase.key.id())

  policy = issue_tracker_policy.get(issue_tracker.project)
  is_crash = not utils.sub_string_exists_in(NON_CRASH_TYPES,
                                            testcase.crash_type)
  properties = policy.get_new_issue_properties(
      is_security=testcase.security_flag, is_crash=is_crash)

  issue = issue_tracker.new_issue()
  issue.title = data_handler.get_issue_summary(testcase)
  issue.body = data_handler.get_issue_description(
      testcase, reporter=user_email, show_reporter=True)

  # Add reproducibility flag label.
  if testcase.one_time_crasher_flag:
    issue.labels.add(policy.label('unreproducible'))
  else:
    issue.labels.add(policy.label('reproducible'))

  # Chromium-specific labels.
  if issue_tracker.project == 'chromium' and testcase.security_flag:
    # Add reward labels if this is from an external fuzzer contribution.
    fuzzer = data_types.Fuzzer.query(
        data_types.Fuzzer.name == testcase.fuzzer_name).get()
    if fuzzer and fuzzer.external_contribution:
      issue.labels.add('reward-topanel')
      issue.labels.add('External-Fuzzer-Contribution')

    update_issue_impact_labels(testcase, issue)

  # Add additional labels from the job definition and fuzzer.
  additional_labels = data_handler.get_additional_values_for_variable(
      'AUTOMATIC_LABELS', testcase.job_type, testcase.fuzzer_name)
  for label in additional_labels:
    issue.labels.add(label)

  # Add additional components from the job definition and fuzzer.
  automatic_components = data_handler.get_additional_values_for_variable(
      'AUTOMATIC_COMPONENTS', testcase.job_type, testcase.fuzzer_name)
  for component in automatic_components:
    issue.components.add(component)

  # Add issue assignee from the job definition and fuzzer.
  automatic_assignee = data_handler.get_additional_values_for_variable(
      'AUTOMATIC_ASSIGNEE', testcase.job_type, testcase.fuzzer_name)
  if automatic_assignee:
    issue.status = policy.status('assigned')
    issue.assignee = automatic_assignee[0]
  else:
    issue.status = properties.status

  # Add additional ccs from the job definition and fuzzer.
  ccs = data_handler.get_additional_values_for_variable(
      'AUTOMATIC_CCS', testcase.job_type, testcase.fuzzer_name)

  # For externally contributed fuzzers, potentially cc the author.
  # Use fully qualified fuzzer name if one is available.
  fully_qualified_fuzzer_name = (
      testcase.overridden_fuzzer_name or testcase.fuzzer_name)
  ccs += external_users.cc_users_for_fuzzer(fully_qualified_fuzzer_name,
                                            testcase.security_flag)
  ccs += external_users.cc_users_for_job(testcase.job_type,
                                         testcase.security_flag)

  # Add the user as a cc if requested, and any default ccs for this job.
  # Check for additional ccs or labels from the job definition.
  if additional_ccs:
    ccs += [cc for cc in additional_ccs if cc not in ccs]

  # For user uploads, we assume the uploader is interested in the issue.
  if testcase.uploader_email and testcase.uploader_email not in ccs:
    ccs.append(testcase.uploader_email)

  ccs.extend(properties.ccs)

  # Get view restriction rules for the job.
  issue_restrictions = data_handler.get_value_from_job_definition(
      testcase.job_type, 'ISSUE_VIEW_RESTRICTIONS', 'security')
  should_restrict_issue = (
      issue_restrictions == 'all' or
      (issue_restrictions == 'security' and testcase.security_flag))

  has_accountable_people = bool(ccs)

  # Check for labels with special logic.
  additional_labels = []
  if should_restrict_issue:
    additional_labels.append(policy.label('restrict_view'))

  if has_accountable_people:
    additional_labels.append(policy.label('reported'))

  if testcase.security_flag:
    additional_labels.append(policy.label('security_severity'))

  additional_labels.append(policy.label('os'))

  # Apply label substitutions.
  for label in itertools.chain(properties.labels, additional_labels):
    for result in apply_substitutions(policy, label, testcase,
                                      security_severity):
      issue.labels.add(result)

  issue.body += data_handler.format_issue_information(
      testcase, properties.issue_body_footer)
  if (should_restrict_issue and has_accountable_people and
      policy.deadline_policy_message):
    issue.body += '\n\n' + policy.deadline_policy_message

  for cc in ccs:
    issue.ccs.add(cc)

  # Add additional labels and components from testcase metadata.
  metadata_labels = _get_from_metadata(testcase, 'issue_labels')
  for label in metadata_labels:
    issue.labels.add(label)

  metadata_components = _get_from_metadata(testcase, 'issue_components')
  for component in metadata_components:
    issue.components.add(component)

  if testcase.one_time_crasher_flag and policy.unreproducible_component:
    issue.components.add(policy.unreproducible_component)

  issue.reporter = user_email
  issue.save()

  # Update the testcase with this newly created issue.
  testcase.bug_information = str(issue.id)
  testcase.put()

  data_handler.update_group_bug(testcase.group_id)
  return issue.id



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Constants shared across butler commands."""

import collections
import os
import sys

# Chromedriver related constants.
CHROMEDRIVER_VERSION_URL = (
    'https://commondatastorage.googleapis.com/chromedriver/LATEST_RELEASE_91')
CHROMEDRIVER_DOWNLOAD_PATTERN = (
    'https://commondatastorage.googleapis.com/chromedriver/{version}/'
    '{archive_name}')

# Local directory of deployment files.
PACKAGE_TARGET_ZIP_DIRECTORY = 'deployment'

# Deprecated source archive name.
LEGACY_ZIP_NAME = 'clusterfuzz-source.zip'

# File containing the source revision information.
PACKAGE_TARGET_MANIFEST_PATH = os.path.join('src', 'appengine', 'resources',
                                            'clusterfuzz-source.manifest')

# Supported Platforms and ABIS (newer to older order).
PLATFORMS = collections.OrderedDict([
    ('windows', 'win_amd64'),
    ('macos', ('macosx_10_14_x86_64', 'macosx_10_9_x86_64',
               'macosx_10_12_x86_64')),
    ('linux', ('manylinux2010_x86_64', 'manylinux1_x86_64')),
])

if sys.version_info.major == 3 and sys.version_info.minor == 7:
  ABIS = {'linux': 'cp37m', 'windows': 'cp37m', 'macos': 'cp37m'}
elif sys.version_info.major == 3 and sys.version_info.minor == 8:
  ABIS = {'linux': 'cp38', 'windows': 'cp38', 'macos': 'cp38'}
elif sys.version_info.major == 3 and sys.version_info.minor == 9:
  ABIS = {'linux': 'cp39', 'windows': 'cp39', 'macos': 'cp39'}
else:
  raise Exception('Only python versions 3.7-3.9 are supported.')

# Config directory to use for tests.
TEST_CONFIG_DIR = os.path.join('configs', 'test')

# Application id for local testing.
TEST_APP_ID = 'test-clusterfuzz'
TEST_APP_ID_WITH_DEV_PREFIX = 'dev~' + TEST_APP_ID

DEV_APPSERVER_PORT = 9000
DEV_APPSERVER_HOST = 'localhost:' + str(DEV_APPSERVER_PORT)

CRON_SERVICE_PORT = 9009
CRON_SERVICE_HOST = 'localhost:' + str(CRON_SERVICE_PORT)

DEV_APPSERVER_ADMIN_PORT = 9002

DATASTORE_EMULATOR_PORT = 9004
DATASTORE_EMULATOR_HOST = 'localhost:' + str(DATASTORE_EMULATOR_PORT)

PUBSUB_EMULATOR_PORT = 9006
PUBSUB_EMULATOR_HOST = 'localhost:' + str(PUBSUB_EMULATOR_PORT)

LOCAL_GCS_SERVER_PORT = 9008
LOCAL_GCS_SERVER_HOST = 'http://localhost:' + str(LOCAL_GCS_SERVER_PORT)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generate strategy pool for launcher based on strategy probabilities.

Decides the set strategies to be considered by the launcher. Note
that because of compatibility issues, the exact set of strategies
generated here may be modified in the launcher before being launched."""

from collections import namedtuple

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.fuzzing import strategy
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

GENERATORS = [
    strategy.CORPUS_MUTATION_RADAMSA_STRATEGY,
    strategy.CORPUS_MUTATION_ML_RNN_STRATEGY,
]

StrategyCombination = namedtuple('StrategyCombination',
                                 'strategy_name probability')


class StrategyPool(object):
  """Object used to keep track of which strategies the launcher should attempt
  to enable."""

  def __init__(self):
    """Empty set representing empty strategy pool."""
    self.strategy_names = set()

  def add_strategy(self, strategy_tuple):
    """Add a strategy into our existing strategy pool unless it is disabled."""
    if strategy_tuple.name not in environment.get_value('DISABLED_STRATEGIES',
                                                        '').split(','):
      self.strategy_names.add(strategy_tuple.name)

  def do_strategy(self, strategy_tuple):
    """Boolean value representing whether or not a strategy is in our strategy
    pool."""
    return strategy_tuple.name in self.strategy_names


def choose_generator(strategy_pool):
  """Chooses whether to use radamsa, ml rnn, or no generator and updates the
  strategy pool."""

  radamsa_prob = engine_common.get_strategy_probability(
      strategy.CORPUS_MUTATION_RADAMSA_STRATEGY.name,
      default=strategy.CORPUS_MUTATION_RADAMSA_STRATEGY.probability)

  ml_rnn_prob = engine_common.get_strategy_probability(
      strategy.CORPUS_MUTATION_ML_RNN_STRATEGY.name,
      default=strategy.CORPUS_MUTATION_ML_RNN_STRATEGY.probability)

  if engine_common.decide_with_probability(radamsa_prob + ml_rnn_prob):
    if engine_common.decide_with_probability(
        radamsa_prob / (radamsa_prob + ml_rnn_prob)):
      strategy_pool.add_strategy(strategy.CORPUS_MUTATION_RADAMSA_STRATEGY)
    else:
      strategy_pool.add_strategy(strategy.CORPUS_MUTATION_ML_RNN_STRATEGY)


def do_strategy(strategy_tuple):
  """Return whether or not to use a given strategy."""
  return engine_common.decide_with_probability(
      engine_common.get_strategy_probability(strategy_tuple.name,
                                             strategy_tuple.probability))


def generate_default_strategy_pool(strategy_list, use_generator):
  """Return a strategy pool representing a selection of strategies for launcher
  to consider.

  Select strategies according to default strategy selection method."""
  pool = StrategyPool()

  # If use_generator is enabled, decide whether to include radamsa, ml rnn,
  # or no generator (mutually exclusive).
  if use_generator:
    choose_generator(pool)

  # Decide whether or not to add non-generator strategies according to
  # probability parameters.
  for value in [
      strategy_entry for strategy_entry in strategy_list
      if strategy_entry not in GENERATORS
  ]:
    if do_strategy(value):
      pool.add_strategy(value)

  logs.log('Strategy pool was generated according to default parameters. '
           'Chosen strategies: ' + ', '.join(pool.strategy_names))
  return pool


def generate_weighted_strategy_pool(strategy_list, use_generator, engine_name):
  """Generate a strategy pool based on probability distribution from multi armed
  bandit experimentation."""

  # If weighted strategy selection is enabled, there will be a distribution
  # stored in the environment.
  distribution = environment.get_value('STRATEGY_SELECTION_DISTRIBUTION')
  selection_method = environment.get_value(
      'STRATEGY_SELECTION_METHOD', default_value='default')

  # Otherwise if weighted strategy selection is not enabled (strategy selection
  # method is default) or if we cannot query properly, generate strategy
  # pool according to default parameters. We pass the combined list of
  # multi-armed bandit strategies and manual strategies for consideration in
  # the default strategy selection process.
  if not distribution or selection_method == 'default':
    return generate_default_strategy_pool(strategy_list, use_generator)

  # Change the distribution to a list of named tuples rather than a list of
  # dictionaries so that we can use the random_weighted_choice function. Filter
  # out probability entries from other engines.
  distribution_tuples = [
      StrategyCombination(
          strategy_name=elem['strategy_name'], probability=elem['probability'])
      for elem in distribution
      if elem['engine'] == engine_name
  ]

  if not distribution_tuples:
    logs.log_warn('Tried to generate a weighted strategy pool, but do not have '
                  'strategy probabilities for %s fuzzing engine.' % engine_name)
    return generate_default_strategy_pool(strategy_list, use_generator)

  strategy_selection = utils.random_weighted_choice(distribution_tuples,
                                                    'probability')
  strategy_name = strategy_selection.strategy_name

  chosen_strategies = strategy_name.split(',')
  pool = StrategyPool()

  for strategy_tuple in strategy_list:
    if strategy_tuple.name in chosen_strategies:
      pool.add_strategy(strategy_tuple)

  # We consider certain strategies separately as those are only supported by a
  # small number of fuzz targets and should be used heavily when available.
  for value in [
      strategy_entry for strategy_entry in strategy_list
      if strategy_entry.manually_enable
  ]:
    if do_strategy(value):
      pool.add_strategy(value)

  logs.log('Strategy pool was generated according to weighted distribution. '
           'Chosen strategies: ' + ', '.join(pool.strategy_names))
  return pool



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""crash_analyzer tests."""

import os
import unittest

from clusterfuzz._internal.crash_analysis import crash_analyzer
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers


class IgnoreStacktraceTest(unittest.TestCase):
  """Tests CrashComparer."""

  def setUp(self):
    test_helpers.patch_environ(self)

  def test_search_excludes(self):
    """Test SEARCH_EXCLUDES env var works."""
    crash_stacktrace = ('aaa\nbbbbbbb\nccc\nddd\n\n')
    self.assertFalse(crash_analyzer.ignore_stacktrace(crash_stacktrace))

    os.environ['SEARCH_EXCLUDES'] = r'eeee'
    self.assertFalse(crash_analyzer.ignore_stacktrace(crash_stacktrace))

    os.environ['SEARCH_EXCLUDES'] = r'ccc'
    self.assertTrue(crash_analyzer.ignore_stacktrace(crash_stacktrace))

  def test_stack_blacklist_regexes(self):
    """Test stacktrace.stack_blacklist_regexes in project.yaml works."""

    def _mock_config_get(_, param):
      """Handle test configuration options."""
      if param == 'stacktrace.stack_blacklist_regexes':
        return [r'.*[c]{3}']
      return None

    test_helpers.patch(
        self, ['clusterfuzz._internal.config.local_config.ProjectConfig.get'])
    self.mock.get.side_effect = _mock_config_get

    crash_stacktrace = ('aaa\nbbbbbbb\nzzzccc\nddd\n\n')
    self.assertTrue(crash_analyzer.ignore_stacktrace(crash_stacktrace))

    crash_stacktrace = ('aaa\nbbbbbbb\nddd\n\n')
    self.assertFalse(crash_analyzer.ignore_stacktrace(crash_stacktrace))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for process management."""

import copy
import datetime
import os
import queue
import subprocess
import sys
import threading
import time

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.crash_analysis import crash_analyzer
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.platforms import android
from clusterfuzz._internal.platforms import linux
from clusterfuzz._internal.platforms import windows
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell

# FIXME: Find a better way to handle this case. These imports
# will fail and are not needed from App Engine.
try:
  import multiprocessing

  import mozprocess
  import psutil
except ImportError:
  pass

# On Android, we need to wait a little after a crash occurred to get the full
# logcat output. This makes sure we get all the stack frames since there is no
# effective end marker.
ANDROID_CRASH_LOGCAT_WAIT_TIME = 0.3

# Time in seconds it usually takes to analyze a crash. This is usually high
# in case of Android where it is required to do several adb shell calls.
CRASH_ANALYSIS_TIME = 1.5

# Test timeout if not specified.
DEFAULT_TEST_TIMEOUT = 10

# Time to wait for cleanup after process if finished.
PROCESS_CLEANUP_WAIT_TIME = 5

# LeakSanitizer needs additional time to process all leaks and dump stacks on
# process shutdown.
LSAN_ANALYSIS_TIME = 1

# Time to wait for thread cleanup (e.g. dumping coverage, etc).
THREAD_FINISH_WAIT_TIME = 5


class ProcessStatus(object):
  """Process exited notification."""

  def __init__(self):
    self.finished = False

  def __call__(self):
    self.finished = True


def start_process(process_handle):
  """Start the process using process handle and override list2cmdline for
  Windows."""
  is_win = environment.platform() == 'WINDOWS'
  if is_win:
    # Override list2cmdline on Windows to return first index of list as string.
    # This is to workaround a mozprocess bug since it passes command as list
    # and not as string.
    subprocess.list2cmdline_orig = subprocess.list2cmdline
    subprocess.list2cmdline = lambda s: s[0]
  try:
    process_handle.run()
  finally:
    if is_win:
      subprocess.list2cmdline = subprocess.list2cmdline_orig


def cleanup_defunct_processes():
  """Cleans up defunct processes."""
  # Defunct processes happen only on unix platforms.
  if environment.platform() != 'WINDOWS':
    while True:
      try:
        # Matches any defunct child process.
        p, _ = os.waitpid(-1, os.WNOHANG)
        if not p:
          break

        logs.log('Clearing defunct process %s.' % str(p))
      except:
        break


# Note: changes to this function may require changes to untrusted_runner.proto.
# This should only be used for running target black box applications which
# return text output.
def run_process(cmdline,
                current_working_directory=None,
                timeout=DEFAULT_TEST_TIMEOUT,
                need_shell=False,
                gestures=None,
                env_copy=None,
                testcase_run=True,
                ignore_children=True):
  """Executes a process with a given command line and other parameters."""
  if environment.is_trusted_host() and testcase_run:
    from clusterfuzz._internal.bot.untrusted_runner import remote_process_host
    return remote_process_host.run_process(
        cmdline, current_working_directory, timeout, need_shell, gestures,
        env_copy, testcase_run, ignore_children)

  if gestures is None:
    gestures = []

  if env_copy:
    os.environ.update(env_copy)

  # FIXME(mbarbella): Using LAUNCHER_PATH here is error prone. It forces us to
  # do certain operations before fuzzer setup (e.g. bad build check).
  launcher = environment.get_value('LAUNCHER_PATH')

  # This is used when running scripts on native linux OS and not on the device.
  # E.g. running a fuzzer to generate testcases or launcher script.
  plt = environment.platform()
  runs_on_device = environment.is_android(plt) or plt == 'FUCHSIA'
  if runs_on_device and (not testcase_run or launcher):
    plt = 'LINUX'

  is_android = environment.is_android(plt)

  # Lower down testcase timeout slightly to account for time for crash analysis.
  timeout -= CRASH_ANALYSIS_TIME

  # LeakSanitizer hack - give time for stdout/stderr processing.
  lsan = environment.get_value('LSAN', False)
  if lsan:
    timeout -= LSAN_ANALYSIS_TIME

  # Initialize variables.
  adb_output = None
  process_output = ''
  process_status = None
  return_code = 0
  process_poll_interval = environment.get_value('PROCESS_POLL_INTERVAL', 0.5)
  start_time = time.time()
  watch_for_process_exit = (
      environment.get_value('WATCH_FOR_PROCESS_EXIT') if is_android else True)
  window_list = []

  # Get gesture start time from last element in gesture list.
  gestures = copy.deepcopy(gestures)
  if gestures and gestures[-1].startswith('Trigger'):
    gesture_start_time = int(gestures[-1].split(':')[1])
    gestures.pop()
  else:
    gesture_start_time = timeout // 2

  if is_android:
    # Clear the log upfront.
    android.logger.clear_log()

    # Run the app.
    adb_output = android.adb.run_command(cmdline, timeout=timeout)
  else:
    cmd = shell.get_command(cmdline)

    process_output = mozprocess.processhandler.StoreOutput()
    process_status = ProcessStatus()
    try:
      process_handle = mozprocess.ProcessHandlerMixin(
          cmd,
          args=None,
          cwd=current_working_directory,
          shell=need_shell,
          processOutputLine=[process_output],
          onFinish=[process_status],
          ignore_children=ignore_children)
      start_process(process_handle)
    except:
      logs.log_error('Exception occurred when running command: %s.' % cmdline)
      return None, None, ''

  while True:
    time.sleep(process_poll_interval)

    # Run the gestures at gesture_start_time or in case we didn't find windows
    # in the last try.
    if (gestures and time.time() - start_time >= gesture_start_time and
        not window_list):
      # In case, we don't find any windows, we increment the gesture start time
      # so that the next check is after 1 second.
      gesture_start_time += 1

      if plt == 'LINUX':
        linux.gestures.run_gestures(gestures, process_handle.pid,
                                    process_status, start_time, timeout,
                                    window_list)
      elif plt == 'WINDOWS':
        windows.gestures.run_gestures(gestures, process_handle.pid,
                                      process_status, start_time, timeout,
                                      window_list)
      elif is_android:
        android.gestures.run_gestures(gestures, start_time, timeout)

        # TODO(mbarbella): We add a fake window here to prevent gestures on
        # Android from getting executed more than once.
        window_list = ['FAKE']

    if time.time() - start_time >= timeout:
      break

    # Collect the process output.
    output = (
        android.logger.log_output()
        if is_android else b'\n'.join(process_output.output))
    output = utils.decode_to_unicode(output)
    if crash_analyzer.is_memory_tool_crash(output):
      break

    # Check if we need to bail out on process exit.
    if watch_for_process_exit:
      # If |watch_for_process_exit| is set, then we already completed running
      # our app launch command. So, we can bail out.
      if is_android:
        break

      # On desktop, we bail out as soon as the process finishes.
      if process_status and process_status.finished:
        # Wait for process shutdown and set return code.
        process_handle.wait(timeout=PROCESS_CLEANUP_WAIT_TIME)
        break

  # Process output based on platform.
  if is_android:
    # Get current log output. If device is in reboot mode, logcat automatically
    # waits for device to be online.
    time.sleep(ANDROID_CRASH_LOGCAT_WAIT_TIME)
    output = android.logger.log_output()

    if android.constants.LOW_MEMORY_REGEX.search(output):
      # If the device is low on memory, we should force reboot and bail out to
      # prevent device from getting in a frozen state.
      logs.log('Device is low on memory, rebooting.', output=output)
      android.adb.hard_reset()
      android.adb.wait_for_device()

    elif android.adb.time_since_last_reboot() < time.time() - start_time:
      # Check if a reboot has happened, if yes, append log output before reboot
      # and kernel logs content to output.
      log_before_last_reboot = android.logger.log_output_before_last_reboot()
      kernel_log = android.adb.get_kernel_log_content()
      output = '%s%s%s%s%s' % (
          log_before_last_reboot, utils.get_line_seperator('Device rebooted'),
          output, utils.get_line_seperator('Kernel Log'), kernel_log)
      # Make sure to reset SE Linux Permissive Mode. This can be done cheaply
      # in ~0.15 sec and is needed especially between runs for kernel crashes.
      android.adb.run_as_root()
      android.settings.change_se_linux_to_permissive_mode()
      return_code = 1

    # Add output from adb to the front.
    if adb_output:
      output = '%s\n\n%s' % (adb_output, output)

    # Kill the application if it is still running. We do this at the end to
    # prevent this from adding noise to the logcat output.
    task_name = environment.get_value('TASK_NAME')
    child_process_termination_pattern = environment.get_value(
        'CHILD_PROCESS_TERMINATION_PATTERN')
    if task_name == 'fuzz' and child_process_termination_pattern:
      # In some cases, we do not want to terminate the application after each
      # run to avoid long startup times (e.g. for chrome). Terminate processes
      # matching a particular pattern for light cleanup in this case.
      android.adb.kill_processes_and_children_matching_name(
          child_process_termination_pattern)
    else:
      # There is no special termination behavior. Simply stop the application.
      android.app.stop()

  else:
    # Get the return code in case the process has finished already.
    # If the process hasn't finished, return_code will be None which is what
    # callers expect unless the output indicates a crash.
    return_code = process_handle.poll()

    # If the process is still running, then terminate it.
    if not process_status.finished:
      launcher_with_interpreter = shell.get_execute_command(
          launcher) if launcher else None
      if (launcher_with_interpreter and
          cmdline.startswith(launcher_with_interpreter)):
        # If this was a launcher script, we KILL all child processes created
        # except for APP_NAME.
        # It is expected that, if the launcher script terminated normally, it
        # cleans up all the child processes it created itself.
        terminate_root_and_child_processes(process_handle.pid)
      else:
        try:
          # kill() here actually sends SIGTERM on posix.
          process_handle.kill()
        except:
          pass

    if lsan:
      time.sleep(LSAN_ANALYSIS_TIME)

    output = b'\n'.join(process_output.output)
    output = utils.decode_to_unicode(output)

    # X Server hack when max client reached.
    if ('Maximum number of clients reached' in output or
        'Unable to get connection to X server' in output):
      logs.log_error('Unable to connect to X server, exiting.')
      os.system('sudo killall -9 Xvfb blackbox >/dev/null 2>&1')
      sys.exit(0)

  if testcase_run and (crash_analyzer.is_memory_tool_crash(output) or
                       crash_analyzer.is_check_failure_crash(output)):
    return_code = 1

  # If a crash is found, then we add the memory state as well.
  if return_code and is_android:
    ps_output = android.adb.get_ps_output()
    if ps_output:
      output += utils.get_line_seperator('Memory Statistics')
      output += ps_output

  if return_code:
    logs.log_warn(
        'Process (%s) ended with exit code (%s).' % (repr(cmdline),
                                                     str(return_code)),
        output=output)

  return return_code, round(time.time() - start_time, 1), output


def cleanup_stale_processes():
  """Kill stale processes left behind by a job."""
  terminate_multiprocessing_children()
  terminate_stale_application_instances()
  cleanup_defunct_processes()


def close_queue(queue_to_close):
  """Close the queue."""
  if environment.is_trusted_host():
    # We don't use multiprocessing.Queue on trusted hosts.
    return

  try:
    queue_to_close.close()
  except:
    logs.log_error('Unable to close queue.')


def get_process():
  """Return a multiprocessing process object (with bug fixes)."""
  if environment.is_trusted_host():
    # forking/multiprocessing is unsupported because of the RPC connection.
    return threading.Thread

  # FIXME(unassigned): Remove this hack after real bug is fixed.
  # pylint: disable=protected-access
  multiprocessing.current_process()._identity = ()

  return multiprocessing.Process


def get_runtime_snapshot():
  """Return a list of current processes and their command lines as string."""
  process_strings = []
  for process in psutil.process_iter():
    try:
      process_info = process.as_dict(attrs=['name', 'cmdline', 'pid', 'ppid'])
      process_string = '{name} ({pid}, {ppid})'.format(
          name=process_info['name'],
          pid=process_info['pid'],
          ppid=process_info['ppid'])
      process_cmd_line = process_info['cmdline']
      if process_cmd_line:
        process_string += ': {cmd_line}'.format(
            cmd_line=(' '.join(process_cmd_line)))
      process_strings.append(process_string)
    except (psutil.AccessDenied, psutil.NoSuchProcess, OSError):
      # Ignore the error, use whatever info is available for access.
      pass

  return '\n'.join(sorted(process_strings))


def get_queue():
  """Return a multiprocessing queue object."""
  if environment.is_trusted_host():
    # We don't use multiprocessing.Process on trusted hosts. No need to use
    # multiprocessing.Queue.
    return queue.Queue()

  try:
    result_queue = multiprocessing.Queue()
  except:
    # FIXME: Invalid cross-device link error. Happens sometimes with
    # chroot jobs even though /dev/shm and /run/shm are mounted.
    logs.log_error('Unable to get multiprocessing queue.')
    return None

  return result_queue


def terminate_hung_threads(threads):
  """Terminate hung threads."""
  start_time = time.time()
  while time.time() - start_time < THREAD_FINISH_WAIT_TIME:
    if not any([thread.is_alive() for thread in threads]):
      # No threads are alive, so we're done.
      return
    time.sleep(0.1)

  logs.log_warn('Hang detected.', snapshot=get_runtime_snapshot())

  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import host

    # Bail out on trusted hosts since we're using threads and can't clean up.
    host.host_exit_no_return()

  # Terminate all threads that are still alive.
  try:
    [thread.terminate() for thread in threads if thread.is_alive()]
  except:
    pass


def terminate_root_and_child_processes(root_pid):
  """Terminate the root process along with any children it spawned."""
  app_name = environment.get_value('APP_NAME')
  direct_children = utils.get_process_ids(root_pid, recursive=False)

  for child_pid in direct_children:
    # utils.get_process_ids also returns the parent pid.
    if child_pid == root_pid:
      continue

    try:
      child = psutil.Process(child_pid)
    except Exception:
      # Process doesn't exist anymore.
      continue

    if child.name() == app_name:
      # Send SIGTERM to the root APP_NAME process only, and none of its children
      # so that coverage data will be dumped properly (e.g. the browser process
      # of chrome).
      # TODO(ochang): Figure out how windows coverage is dumped since there is
      # no equivalent of SIGTERM.
      terminate_process(child_pid, kill=False)
      continue

    child_and_grand_children_pids = utils.get_process_ids(
        child_pid, recursive=True)
    for pid in child_and_grand_children_pids:
      terminate_process(pid, kill=True)

  terminate_process(root_pid, kill=True)


def terminate_multiprocessing_children():
  """Terminate all children created with multiprocessing module."""
  child_list = multiprocessing.active_children()
  for child in child_list:
    try:
      child.terminate()
    except:
      # Unable to terminate multiprocessing child or was not needed.
      pass


def terminate_stale_application_instances():
  """Kill stale instances of the application running for this command."""
  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import remote_process_host
    remote_process_host.terminate_stale_application_instances()
    return

  # Stale instance cleanup is sometimes disabled for local testing.
  if not environment.get_value('KILL_STALE_INSTANCES', True):
    return

  additional_process_to_kill = environment.get_value(
      'ADDITIONAL_PROCESSES_TO_KILL')
  builds_directory = environment.get_value('BUILDS_DIR')
  llvm_symbolizer_filename = environment.get_executable_filename(
      'llvm-symbolizer')
  platform = environment.platform()
  start_time = time.time()

  processes_to_kill = []
  # Avoid killing the test binary when running the reproduce tool. It is
  # commonly in-use on the side on developer workstations.
  if not environment.get_value('REPRODUCE_TOOL'):
    app_name = environment.get_value('APP_NAME')
    processes_to_kill += [app_name]

  if additional_process_to_kill:
    processes_to_kill += additional_process_to_kill.split(' ')
  processes_to_kill = [x for x in processes_to_kill if x]

  if environment.is_android(platform):
    # Cleanup any stale adb connections.
    device_serial = environment.get_value('ANDROID_SERIAL')
    adb_search_string = 'adb -s %s' % device_serial

    # Terminate llvm symbolizer processes matching exact path. This is important
    # for Android where multiple device instances run on same host.
    llvm_symbolizer_path = environment.get_llvm_symbolizer_path()

    terminate_processes_matching_cmd_line(
        [adb_search_string, llvm_symbolizer_path], kill=True)

    # Make sure device is online and rooted.
    android.adb.run_as_root()

    # Make sure to reset SE Linux Permissive Mode (might be lost in reboot).
    android.settings.change_se_linux_to_permissive_mode()

    # Make sure that device forwarder is running (might be lost in reboot or
    # process crash).
    android.device.setup_host_and_device_forwarder_if_needed()

    # Make sure that package optimization is complete (might be triggered due to
    # unexpected circumstances).
    android.app.wait_until_optimization_complete()

    # Reset application state, which kills its pending instances and re-grants
    # the storage permissions.
    android.app.reset()

  elif platform == 'WINDOWS':
    processes_to_kill += [
        'cdb.exe',
        'handle.exe',
        'msdt.exe',
        'openwith.exe',
        'WerFault.exe',
        llvm_symbolizer_filename,
    ]
    terminate_processes_matching_names(processes_to_kill, kill=True)
    terminate_processes_matching_cmd_line(builds_directory, kill=True)

    # Artifical sleep to let the processes get terminated.
    time.sleep(1)

  else:
    # Handle Linux and Mac platforms.
    processes_to_kill += [
        'addr2line',
        'atos',
        'chrome-devel-sandbox',
        'gdb',
        'nacl_helper',
        'xdotool',
        llvm_symbolizer_filename,
    ]
    terminate_processes_matching_names(processes_to_kill, kill=True)
    terminate_processes_matching_cmd_line(builds_directory, kill=True)

  duration = int(time.time() - start_time)
  if duration >= 5:
    logs.log('Process kill took longer than usual - %s.' % str(
        datetime.timedelta(seconds=duration)))


def terminate_process(process_id, kill=False):
  """Terminates a process by its process id."""
  try:
    process = psutil.Process(process_id)

    if kill:
      process.kill()
    else:
      process.terminate()

  except (psutil.AccessDenied, psutil.NoSuchProcess, OSError):
    logs.log_warn('Failed to terminate process.')


def terminate_processes_matching_names(match_strings, kill=False):
  """Terminates processes matching particular names (case sensitive)."""
  if isinstance(match_strings, str):
    match_strings = [match_strings]

  for process in psutil.process_iter():
    try:
      process_info = process.as_dict(attrs=['name', 'pid'])
      process_name = process_info['name']
    except (psutil.AccessDenied, psutil.NoSuchProcess, OSError):
      continue

    if any(x == process_name for x in match_strings):
      terminate_process(process_info['pid'], kill)


def terminate_processes_matching_cmd_line(match_strings,
                                          kill=False,
                                          exclude_strings=None):
  """Terminates processes matching particular command line (case sensitive)."""
  if exclude_strings is None:
    # By default, do not terminate processes containing butler.py. This is
    # important so that the reproduce tool does not terminate itself, as the
    # rest of its command line may contain strings we usually terminate such
    # as paths to build directories.
    exclude_strings = ['butler.py', 'reproduce.sh']

  if isinstance(match_strings, str):
    match_strings = [match_strings]

  for process in psutil.process_iter():
    try:
      process_info = process.as_dict(attrs=['cmdline', 'pid'])
      process_cmd_line = process_info['cmdline']
      if not process_cmd_line:
        continue
      process_path = ' '.join(process_cmd_line)
    except (psutil.AccessDenied, psutil.NoSuchProcess, OSError):
      continue

    if any(x in process_path for x in match_strings):
      if not any([x in process_path for x in exclude_strings]):
        terminate_process(process_info['pid'], kill)
